#!/bin/sh

#*************************************************#
#              build_iscsi_qla4x00.sh             #
#	       QLogic Corporation		  #
#               May 19, 2005	                  #
#                                                 #
#  Build script for qla4xxx driver on 2.4 Kernel  #
#     Adapted from Dell's DKMS pkg		  #
#     Licensed under the GNU GPL		  #
#*************************************************#

DRIVER="qla4xxx"
QLOGIC="qlogic"
MODULE_NAME=( "qla4010" "qla4022" )
RPMVERSION=v3.22-2
# Install source in /usr/src/qlogic/<driver ver-rpm release>/
QL_SOURCE=/usr/src/$QLOGIC/$RPMVERSION
QL_BACKUP=$QL_SOURCE/backup
QL_CONF_ENTRIES=$QL_BACKUP/ql_conf_entries
QL_KERNEL_ENTRIES=$QL_BACKUP/ql_kernel_entries # foe SuSE /etc/sysconfig/kernel
QL_MOD_REMOVED=$QL_BACKUP/ql_mod_removed

mkdir -p $QL_SOURCE
mkdir -p $QL_BACKUP

source $QL_SOURCE/ql_common.sh

# Updates the ramdisk
function update_initrd() {
	create_ramdisk "${MODULE_NAME[@]}"
	return $?
}

install_api() {
        # Install the SNIA API
        #echo -e "Installing the QLogic SNIA API library libqlsdm.so in /usr/lib."
        #echo -e "API package is installed in the $QL_SOURCE/ApiPkg path."
        cd $QL_SOURCE
        ./libinstall
}

# --------------------------------------------------------- #
# full_install ()	                                    #
# This is full installation from building the driver, 	    #
# intalling it in /lib/modules/`uname -r` to creating the   #
# ramdisk						    #
#							    #
# Parameter: None					    # 
# Returns: 0 on success, 1 on fail			    #
# --------------------------------------------------------- #
function full_install()
{
	# Build kernel dependency if needed
	kernel_dep
	# Build the driver
	build_driver
	# Check if build was successful
	if [ $? != 0 ] || [ ! -e "qla4010.o" ]; then # check for qla4010.o only
		echo "# $KERNEL_VER install failed.  Remove all references to this kernel in" >> $INSTALL_LOG
		echo "# this file and reboot your kernel in order to attempt install again." >> $INSTALL_LOG
		echo -e "\nInstallation of ${DRIVER} for $KERNEL_VER failed." 
		echo -e "Check $QL_SOURCE/make.$KERNEL_VER.$RPMVERSION.log for make failures." 
		$LOGGER -t qlogic "Installation of ${DRIVER} driver  for $KERNEL_VER failed." \
				  " Check $QL_SOURCE/make.$KERNEL_VER.log for make failures."
		exit 1
	fi

	install_driver
	if [ $? != 0 ]; then
		# something went wrong
		echo "Driver installation failed"
		echo "Restoring original driver, ramdisk and modules.conf"
		restore_original
		exit 1
	fi


	# Now called from top level script
	#update_initrd

	#install_api

	if [ $? = 0 ]; then
		# Need to reboot, during future reboots it will automatically be loaded
		$LOGGER -t qlogic "Installation of ${DRIVER} driver for $KERNEL_VER successful."
		exit 0
	fi
	exit 1
}

# START
if [ "$DISTRO" = "" ]; then 
    echo -e "Distribution cannot be determined"
    echo -e "Note:This Source RPM  supports driver installation 
	     for RedHat, SuSE and Asianux distribution only"
    exit 1
fi

if [ "`uname -m`"  = "ia64" ]; then
	if [ "$DISTRO" = "$REDHAT" ]; then 
	    BOOTDIR="/boot/efi/efi/redhat"
	elif [ "$DISTRO" = "$ASIANUX" ]; then 
	    BOOTDIR="/boot/efi/efi/asianux"
	fi
fi

if [ "$DISTRO" = "$REDHAT" ] || [ "$DISTRO" = "$ASIANUX" ]; then
	LOGGER="/usr/bin/logger"
else
	LOGGER="/bin/logger"
fi

#check_kernel_source

cd $QL_SOURCE

if ! [ -e $INSTALL_LOG ] || [ `grep -c "^$KERNEL_VER$" $INSTALL_LOG 2>/dev/null` -eq 0 ]; then
   # echo -e "Installing ${MODULE_NAME} driver for $KERNEL_VER..." 

    # Check if install_log needs to be created
    if ! [ -e $INSTALL_LOG ]; then #double check, need to eliminate
	echo -e ""
	echo -e "Creating initial $QL_SOURCE/$INSTALL_LOG..." 
	echo -e "# This file is automatically generated as part of the ${DRIVER} RPM"	>> $INSTALL_LOG
    fi
fi

if [ $# = 0 ]; then
# Default option, install
	full_install
else
	case $1 in
		-i | --install )
			# install the driver
			full_install
		;;
		-d | --ramdisk )
			# create the ramdisk
			create_ramdisk "$2" "$3" "$4" 
			exit $?
		;;
                -r | --restore )
                        # resoter original
                        restore_original
                ;;

		* )
			# invalid option
			echo "Invalid option $1"
		;;
	esac

fi
#END
