#*************************************************#
#                   ql_common.sh                  #
#		QLogic Corporation		  #
#                February 01, 2005                #
#                                                 #
#   Common functions for FC/iSCSI build script	  #
#*************************************************#

KERNEL_VER=`uname -r`
INSTALL_LOG="install.$RPMVERSION.log"
SOURCEDIR="/lib/modules/$KERNEL_VER/build"
#Distribution specific
SCSIDIR="/lib/modules/$KERNEL_VER/kernel/drivers/scsi"
SCSI_UPDATE_DIR="/lib/modules/$KERNEL_VER/updates"
mkdir -p "$SCSIDIR"
USERSRC=/usr/src
#Kernel EXTRAVERSION
EXT=""
LOGGER=""
BOOTDIR="/boot"
MOD_FILE=modules.conf
MODCONF="/etc/$MOD_FILE"
SCSIHOSTADAPTER="scsi_hostadapter"
#For RedHat distribution increase the ram disk size.
MKINITRD="/sbin/mkinitrd"
SYS_FILE="kernel"
SYSCONFIG="/etc/sysconfig/$SYS_FILE"
NO_BACKUP=1
REBUILD=1

#Distributions
SUSE="SUSE"
REDHAT="REDHAT"
ASIANUX="ASIANUX"
DISTRO=""
if [ -f "/etc/SuSE-release" ]; then
	DISTRO=$SUSE
elif [ -f "/etc/asianux-release" ]; then
	DISTRO=$ASIANUX
elif [ -f "/etc/redhat-release" ]; then
	DISTRO=$REDHAT
fi
# Processing is same for Red Hat and Asianux



# --------------------------------------------------------- #
# invoke_cmd ()		                                    #
# Invokes the shell command				    #
# Parameter: 	 		                            #
#	$1    the command to be invoked			    #
# 	$2    the text output to the screen                 #
# 	$3    if set to "background", outputs "."s 	    #
#	      while waiting for completion		    #
#							    #
# Returns: Commands exit value.				    #
# --------------------------------------------------------- #
function invoke_cmd() {
    exitval=0
    rm -f .tmp_exitval
    if [ -e .tmp_exitval ]; then
	echo $"invoke_command could not run!" >$std_err
	exit 1
    fi
    if [ "$2" != "" ]; then	
	    echo -en "$2..."
    fi	 
    if [ "$3" == background ]; then
	SLEEP_VAL=.08
	sleep $SLEEP_VAL >& /dev/null
	if [ $? != 0 ]; then
		SLEEP_VAL=1
	fi
	
	(eval $1; echo "exitval=$?" >> .tmp_exitval) &
	while ! [ -e .tmp_exitval ]
	do
	  if [ "$4" == "progress" ]; then
	    #echo -en '\E[34;44m '
		echo -en "|"
		echo -en "\b"
		sleep $SLEEP_VAL
		echo -en "/"
		echo -en "\b"
		sleep $SLEEP_VAL
		echo -en "-"
		echo -en "\b"
		sleep $SLEEP_VAL 
		echo -en "\\"
		echo -en "\b"
		sleep $SLEEP_VAL
	  else
	   echo -en "." 
	   sleep 1
	  fi
	done
	echo -en "\b"
	echo " "
	#tput sgr0
	. .tmp_exitval
	rm .tmp_exitval
    else
	eval $1; exitval=$?
    fi
    if [ $exitval -gt 0 ]; then
	echo -en "(bad exit status)" 
	script_location=`dirname $0`
	echo "$1 failed to execute"
	echo $1 >> $script_location/.invoke_command.log
    fi
    echo -en "\n" 
    return $exitval
}

# --------------------------------------------------------- #
# edit_sysconfig ()	                                    #
# Function to modify list of modules to be included 	    #
# in the Ram Disk for SuSE Distribution. Edits the	    #
# /etc/sysconfig/kernel					    #
#							    #
# Parameter: 	 		                            #
#	$*    Space separated module list		    #
# Returns: None						    #
# --------------------------------------------------------- #
function edit_sysconfig() {
    	MODLIST="$1"
	# check if Module already in the sysconfig/kernel
	grep -w "$MODLIST" $SYSCONFIG >& /dev/null
	if [ $? = 0 ]; then
		# MODLIST already there, quit from here
		return 1
	fi
	MODULES=`cat $SYSCONFIG | grep INITRD_MODULES | 
			awk 'BEGIN {FS="="} {print $2}'| sed 's/\"//g'`
	MODULES="$MODULES $MODLIST"
	echo "New module list: $MODULES"
	sed "s/INITRD_MODULES.*/INITRD_MODULES=\"$MODULES\"/"  $SYSCONFIG>/tmp/kernel
	cp -f /tmp/kernel $SYSCONFIG

	if [ -f $QL_KERNEL_ENTRIES ]; then
		grep w "$MODLIST" $QL_KERNEL_ENTRIES >& /dev/null
		if [ $? = 0 ]; then
			return 0	
		fi
	fi
	echo "$MODLIST" >> $QL_KERNEL_ENTRIES
	return 0
}

# --------------------------------------------------------- #
# add_entry_to_modconf ()                                   #
# Add module entry to /etc/modules.conf			    #
#							    #
# Parameter: 	 		                            #
#	$1    Module name to be added to /etc/module.conf   #
# Returns: None						    #
# --------------------------------------------------------- #
function add_entry_to_modconf() {
	# get the scsi_hostadapterxx number
	DONE=0
	CNT=""

	# check if the entry is already in the /etc/modules.conf
	# if not then add
	grep "^[[:space:]]*alias[[:space:]]\+${SCSIHOSTADAPTER}.*${1}$" ${MODCONF} >& /dev/null 
	if [ $? = 0 ]; then # entry already in /etc/modules.conf
		return 1
	fi

	while [ $DONE -eq 0 ]
	do
 		TEXT=`cat ${MODCONF} | grep "^[[:space:]]*alias[[:space:]]\+${SCSIHOSTADAPTER}${CNT}"`

 		if [ "$TEXT" = "" ]; then
  			DONE=1
 		else
			if [ "$CNT" = "" ]; then
				let CNT=0
			fi
  		CNT=`expr $CNT + 1`
 		fi
    	done

    	#echo "adding line: alias ${SCSIHOSTADAPTER}${CNT} $1 to $MODCONF"
	ENTRY_LINE="alias ${SCSIHOSTADAPTER}${CNT} $1" 
    	echo $ENTRY_LINE >> $MODCONF
	# do not add if entry already in, as we might come back here
	# from the include driver in ramdisk option
	if [ -f $QL_CONF_ENTRIES ]; then
		grep "$ENTRY_LINE" $QL_CONF_ENTRIES >& /dev/null
		if [ $? = 0 ]; then
			return 0	
		fi
	fi
    	echo $ENTRY_LINE >> $QL_CONF_ENTRIES
	return 0
}

# --------------------------------------------------------- #
# rename_driver ()	                                    #
# Rename the old driver if exists and store it		    #
# in the $QL_BACKUP directory				    #
# If there is a backup already then don't do anything	    #
# Keep it as it is					    #
#							    #
# Parameter: 	 		                            #
# 	$1   Module name to be renamed 			    #
# Returns: None						    #
# --------------------------------------------------------- #
function rename_driver() {
	#check if argument provided
	if [ -z "$1" ]; then
		echo "rename_driver: Missing argument-driver name"
		exit 1
	fi
	ORIGINAL_DRIVER=$SCSIDIR/$1.o
	BACKUP_DRIVER=$QL_BACKUP/$1.o-$KERNEL_VER.bak
	
	# Need to backup the update drivers as well
	ORIGINAL_UPDATE_DRIVER=$SCSI_UPDATE_DIR/$1.o
	BACKUP_UPDATE_DRIVER=$SCSI_UPDATE_DIR/$1-old.o


	# backup original driver in 
	# /usr/src/qlogic/<driver version/backup/old_driver.o-<kernel ver>.bak
	if [ ! -e $BACKUP_DRIVER ]; then
		if [ -e $ORIGINAL_DRIVER ]; then
			mv -f $ORIGINAL_DRIVER $BACKUP_DRIVER 
		fi
	fi

	if [ ! -f $BACKUP_UPDATE_DRIVER ]; then
		if [ -f $ORIGINAL_UPDATE_DRIVER ]; then
			mv -f $ORIGINAL_UPDATE_DRIVER $BACKUP_UPDATE_DRIVER 
		fi
	fi
} 

# --------------------------------------------------------- #
# kernel_dep ()		                                    #
# Make kernel dependency 				    #
# Required for SuSE version less than 2.4.21		    #
#							    #
# Parameter: None	 		                    #
# Returns: None						    #
# --------------------------------------------------------- #
function kernel_dep() {
    # Setup correct kernel config before compile
	cd $SOURCEDIR
	echo -e "Please wait: Preparing $DRIVER modular driver build"
	#Check we are 2.4 and lower than 21 on SuSE
	if [ "$DISTRO" = "$SUSE" ]; then 
		if [ "`echo $KERNEL_VER | cut -c 1-3`" = "2.4" ] &&
                   [ "`echo $KERNEL_VER | cut -c 5-6`" -lt 21 ]; then 
			if ! [ -e $SOURCEDIR/include/linux/modversions.h ]; then
				echo "#include <linux/modsetver.h>" > /tmp/modversions.h
			else
				cp  $SOURCEDIR/include/linux/modversions.h /tmp/modversions.h
			fi
			invoke_cmd "make mrproper >/dev/null 2>&1" 
		
			#Copy the /boot/vmlinuz.config to /lib/modules/$KERNEL_VER/build
			echo -e "Note:Assuming the configuration file:$BOOTDIR/vmlinuz.config" 
			echo -e "is for the kernel-$KERNEL_VER.If not then make sure that"
			echo -e "this conf file is for the kernel which you have specified or"
			echo -e "of the running kernel.And run the install command again"
			sleep 4
			invoke_cmd "cp $BOOTDIR/vmlinuz.config .config" \
			"copying $BOOTDIR/vmlinuz.config to  $SOURCEDIR/.config"

			echo -e "Building Kernel Dependencies"	
			invoke_cmd "yes | make  oldconfig >/dev/null 2>&1"
			invoke_cmd "make   dep >/dev/null 2>&1" "" "background" "progress" 
			mv /tmp/modversions.h include/linux/modversions.h
		fi

    	#elif [ ! -f "/usr/src/linux-$KERNEL_VER/include/linux/version.h" ]; then
    	elif [ ! -f "$SOURCEDIR/include/linux/version.h" ]; then
		# Should do this only if 
		# /usr/src/linux-<kernel version>/include/linux/version.h not present
		invoke_cmd "make mrproper >/dev/null 2>&1"
		#copy the correct config file
		CONFIG_FILE="kernel-`echo $KERNEL_VER | sed 's/-.*//'`-`uname -m`"
		#SMP
		if [ `echo $KERNEL_VER | grep smp` ]; then
			CONFIG_FILE="$CONFIG_FILE-smp.config"
		#Enterprise
		elif [ `echo $KERNEL_VER | grep ent` ]; then
			CONFIG_FILE="$CONFIG_FILE-enterprise.config"
		#bigmem
		elif [ `echo $KERNEL_VER | grep bigmem` ]; then
			CONFIG_FILE="$CONFIG_FILE-bigmem.config"
		#default
		else
			CONFIG_FILE="$CONFIG_FILE.config"
		fi

		invoke_cmd "cp ./configs/$CONFIG_FILE .config" " copying $CONFIG_FILE to $SOURCEDIR/.config" 
		echo -e "Building Kernel Dependencies"	
		EXT=`echo $KERNEL_VER | awk 'BEGIN {FS="-"} {print $2}'`
		invoke_cmd "yes | make EXTRAVERSION=-${EXT} oldconfig >/dev/null 2>&1"
		invoke_cmd "make EXTRAVERSION=-${EXT}  dep >/dev/null 2>&1" 
	fi 
}

# --------------------------------------------------------- #
# build_driver ()	                                    #
# Compile the driver dependent on smp or not.		    #
# SuSE has module versioning				    #
# disabled by default in vmlinuz.config			    #
#							    #
# Parameter: None	 		                    #
# Returns: 0 on success, 1 on fail			    #
# --------------------------------------------------------- #
function build_driver() {
	ret_status=0
	cd $QL_SOURCE

	PLATFORM=`uname -m`
	if [ $PLATFORM = 'x86_64' ]; then
		KERNEL_VER=`uname -a`
	fi

	touch *
	make clean >/dev/null
	if [ `echo $KERNEL_VER | grep -i -o smp` ] ||
	   [ `echo $KERNEL_VER | grep ent` ] ||
	   [ `echo $KERNEL_VER | grep bigmem` ]; then
		echo "building for SMP"
		#invoke_cmd "make SMP=1 INCLUDEDIR=$SOURCEDIR/include all  > make.$KERNEL_VER.$RPMVERSION.log 2>&1"
		KERNEL_VER=`uname -r`
		invoke_cmd "make  SMP=1 INCLUDEDIR=$SOURCEDIR/include > make.$KERNEL_VER.$RPMVERSION.log 2>&1" "" "background" "progress"
		ret_status=$?
	else
		echo "building for UP"
		KERNEL_VER=`uname -r`
		#invoke_cmd "make  INCLUDEDIR=$SOURCEDIR/include all > make.$KERNEL_VER.$RPMVERSION.log 2>&1" "" "background" "progress"
		invoke_cmd "make  SMP=0 INCLUDEDIR=$SOURCEDIR/include > make.$KERNEL_VER.$RPMVERSION.log 2>&1" "" "background" "progress"
		ret_status=$?
		echo -e ""
	fi

	return $ret_status;
}

# --------------------------------------------------------- #
# check_kernel_source ()                                    #
# Check to see if kernel source exists			    #
#							    #
# Parameter: None	 		                    #
# Returns: None						    #
# --------------------------------------------------------- #
function check_kernel_source() {
	if ! [ -e $SOURCEDIR/include ]; then
	    echo -e
	    echo -e "Kernel source for $KERNEL_VER kernel is not installed. Please"
	    echo -e "install kernel source from installation CD: "
	    echo -e 	
	    echo -e "For RedHat distribution install RPM -"kernel-source-$KERNEL_VER" "
	    echo -e "from its installation CD under RedHat/RPMS directory by executing"
	    echo -e "the following command:"
	    echo -e "  # rpm -iv kernel-source*.rpm "
	    echo -e 
	    echo -e "For SuSE distribution install the kernel source from installation"
	    echo -e "CD by executing the following command:"
	    echo -e " # yast -i kernel-source" or
	    echo -e " # yast2  -i kernel-source"
	    echo -e
	    echo -e "The directory $SOURCEDIR/include does not exist!"
	    echo -e "Aborting..."
	    echo -e
	    $LOGGER -t qlogic "Installation of QLogic Source RPM could not proceed because" \
			       " kernel source for $KERNEL_VER is not installed." 
	    exit 1
	fi
}

# --------------------------------------------------------- #
# create_ramdisk ()	                                    #
# Adds modules in /etc/modules.conf. Backups current initrd #
# Creates a new initrd with the driver module to be loaded  #
# at boot time						    #
#							    #
# Parameter:						    # 
#	$1	Array of modules to be added 		    #
#		in /etc/modules.conf			    #
#	$2	NO_BACKUP: Do not take backup of initrd	    #
#	$3	FORCE_REBUILD:                              #
#		       "YES": Build ramdisk always	    #
#		       "NO": Build ramdisk only if required #
# Returns: 0 on success, 1 on fail			    #
# --------------------------------------------------------- #
function create_ramdisk() {

	MODULE_LIST=(`echo "$1"`)
	FORCE_REBUILD=$3
	BACKUP_DATE=`date +%m%d%y-%H%M%S`
	SAVING_STR="Saved copy of" 
	ret_status=0
	NEED_REBUILD=0
	BACKUP_FILE_K=""
	BACKUP_FILE_MOD=""
	ORIGINAL_FILE_K=""
	ORIGINAL_FILE_MOD=""

	if [ "$FORCE_REBUILD" = "" ]; then
		FORCE_REBUILD="YES"
	fi

	# LSC: Moved the backup logic to top-level qlinstall script
	# Need to backup sysconfig once
	#if [ "$DISTRO" = "$SUSE" ] && [ ${#MODULE_LIST[@]} -gt 0 ] && [ "$2" != "$NO_BACKUP" ]; then
	#	BACKUP_FILE_K=$QL_BACKUP/$SYS_FILE-$KERNEL_VER-${BACKUP_DATE}.bak
	#	ORIGINAL_FILE_K=$SYSCONFIG 
	#	cp $SYSCONFIG $BACKUP_FILE_K
	#fi	
	
	#backup modules.conf
	#if [ ${#MODULE_LIST[@]} -gt 0 ] && [ "$2" != "$NO_BACKUP" ]; then
	#	BACKUP_FILE_MOD="$QL_BACKUP/$MOD_FILE-$KERNEL_VER-${BACKUP_DATE}.bak"
	#	ORIGINAL_FILE_MOD=$MODCONF 
        #        # backup original modules.conf
        #        cp $MODCONF $BACKUP_FILE_MOD
	#fi

	# mkinitrd
	# Add entry in the /etc/modules.conf
	for MODULE in ${MODULE_LIST[@]}
	do
		MODULE_CONF=""
		if [ "$MODULE" = "2300" ] || [ "$MODULE" = "2200" ]; then
			MODULE_CONF="qla${MODULE}_conf"
		fi
		
		if [ "$DISTRO" = "$SUSE" ]; then
			#SuSe:Default option in .config for scsi_mod is not modular.
			#Append $MODULE modules to the list.
		        if [ "$MODULE_CONF" != "" ]; then
                        	edit_sysconfig "$MODULE_CONF"
				if [ $? = 0 ] ;then
					NEED_REBUILD=1
				fi

                	fi
			edit_sysconfig "qla$MODULE"
			if [ $? = 0 ] ;then
				NEED_REBUILD=1 
			fi
		else #Redhat/Asianux
			if [ "$MODULE_CONF" != "" ]; then
				add_entry_to_modconf "$MODULE_CONF"
				if [ $? = 0 ] ;then
					NEED_REBUILD=1
				fi
			fi 
			add_entry_to_modconf "qla$MODULE"
			if [ $? = 0 ] ;then
				NEED_REBUILD=1
			fi
		fi

	done

	# If force rebuild is not yes then build ramdisk
	# only if NEED_REBUILD is set 	
	#if [ "$FORCE_REBUILD" = "NO" ] && [ $NEED_REBUILD = 0 ]; then
		#do not rebuild 
		#delete any backup created, as it is not required
	#	if [ -f "$BACKUP_FILE_K" ]; then
	#		rm -f $BACKUP_FILE >& /dev/null
	#	fi
	#	if [ -f "$BACKUP_FILE_MOD" ]; then
	#		rm -f $BACKUP_FILE_MOD >& /dev/null
	#	fi
	#	return 1
	#fi

	#if [ -f "$BACKUP_FILE_K" ]; then
	#	# Now give a late message of file saved
        #	echo -e "$SAVING_STR $ORIGINAL_FILE_K as"
        #	echo -e "$BACKUP_FILE_K"
	#	echo -e ""
	#fi
	#if [ -f "$BACKUP_FILE_MOD" ]; then
	#	# Now give a late message of file saved
        #	echo -e "$SAVING_STR $ORIGINAL_FILE_MOD as"
        #	echo -e "$BACKUP_FILE_MOD"
	#	echo -e ""
	#fi

	# should do depmod here and remove from installation function
	# and uninstallaiton part from qla2x00-qlogic.spec 
	echo -e "Building module dependency...."
	# Module dependency
	#invoke_cmd "/sbin/depmod -a -F /boot/System.map-${KERNEL_VER} ${KERNEL_VER}" "depmod" 
	invoke_cmd "/sbin/depmod -a" "depmod" 
	ret_status=$?
	if [ "$ret_status" -ne "0" ]; then
		echo "depmod...failed"
		echo "Please restore backed up files from $QL_BACKUP"
		return $ret_status
	fi

	echo -e ""
	if [ "$ret_status" -eq "0" ]; then
		if [ "$DISTRO" = "$SUSE" ]; then

			if [ "$2" != "$NO_BACKUP" ] && [ -f $BOOTDIR/initrd ]; then
				BACKUP_FILE="$QL_BACKUP/initrd-$KERNEL_VER-$BACKUP_DATE.bak"
				echo -e "$SAVING_STR $BOOTDIR/initrd as"
				echo -e "$BACKUP_FILE"
				invoke_cmd "cp -f $BOOTDIR/initrd $BACKUP_FILE" 
			fi
			echo -e "Creating ramdisk"
			invoke_cmd "/sbin/mk_initrd > /dev/null 2>&1" "mk_initrd" 
			ret_status=$?
			echo ""
		else #Redhat/asianux
			OPTION_SCSI_MOD=`grep "^options scsi_mod" $MODCONF` 
			ret_status=$?
			NEW_OPTIONS="${OPTION_SCSI_MOD}"
			if [ $ret_status -eq 0 ]; then
				#check if max scis luns already there,
				echo "$OPTION_SCSI_MOD" | grep "max_scsi_luns"  >& /dev/null
				if [ $? != 0 ]; then # we need to add
					NEW_OPTIONS="$NEW_OPTIONS max_scsi_luns=255" 
				fi
				#check if allow_ghosts already there
				echo "$OPTION_SCSI_MOD" | grep "scsi_allow_ghost_devices" >& /dev/null 
				if [ $? != 0 ]; then # we need to add
				
					NEW_OPTIONS="$NEW_OPTIONS scsi_allow_ghost_devices=1" 
				else
					#force value to 1
					NEW_OPTIONS=`echo $NEW_OPTIONS | sed "s/scsi_allow_ghost_devices=./scsi_allow_ghost_devices=1/"`
					
				fi	
				if [ "${NEW_OPTIONS}" != "" ]; then
					cat $MODCONF | sed "s/$OPTION_SCSI_MOD/$NEW_OPTIONS/" > /tmp/$MOD_FILE
				if [ -f /tmp/$MOD_FILE ]; then
					mv -f /tmp/$MOD_FILE $MODCONF
				fi

				fi
			else	
				NEW_OPTIONS="options scsi_mod max_scsi_luns=255 scsi_allow_ghost_devices=1"
				echo "${NEW_OPTIONS}" >> $MODCONF
				echo "${NEW_OPTIONS}" >> $QL_CONF_ENTRIES
			fi
			
			if [ "$2" != "$NO_BACKUP" ]; then
				# backup original initrd
				if [ -f $BOOTDIR/initrd-${KERNEL_VER}.img ]; then
					BACKUP_FILE="$QL_BACKUP/initrd-${KERNEL_VER}.img-${BACKUP_DATE}.bak"
					ORIGINAL_FILE="$BOOTDIR/initrd-${KERNEL_VER}.img"
					echo -e "$SAVING_STR $ORIGINAL_FILE as"
					echo -e "$BACKUP_FILE"
					invoke_cmd "cp -f $ORIGINAL_FILE $BACKUP_FILE"
				fi
			fi
			if [ `grep -c "IMAGESIZE.*=[0-3]000" $MKINITRD` -gt  0 ] ; then
				sed "s/IMAGESIZE.*=[0-3]000/IMAGESIZE=4000/" $MKINITRD > /tmp/mkinitrd
				cp -f /tmp/mkinitrd $MKINITRD
			fi
			echo -e "Creating ramdisk"
			if [ -f "$BOOTDIR/initrd-${KERNEL_VER}.img" ]; then
				invoke_cmd "/sbin/mkinitrd -f $BOOTDIR/initrd-${KERNEL_VER}.img ${KERNEL_VER}" \
					"mkinitrd" 
			else
				invoke_cmd "/sbin/mkinitrd $BOOTDIR/initrd-${KERNEL_VER}.img ${KERNEL_VER}" \
					"mkinitrd" 
			fi
			ret_status=$?
		fi
	fi

	if [ "$ret_status" -ne "0" ]; then
		echo "Ramdisk creation failed"
		echo "Please restore backed up files from $QL_BACKUP"
		return $ret_status
	fi

	if [ "$DISTRO" = "$SUSE" ]; then 
		#on SuSE we need to check 
		#/etc/sysconfig/bootloader
		grep "LOADER_TYPE" /etc/sysconfig/bootloader | grep -i "lilo" >& /dev/null
		if [ $? = 0 ]; then
			echo -e "Running LILO...."
			invoke_cmd "/sbin/lilo" "Updating lilo"
		fi
	else	
		# Rerun lilo if necessary
		if [ ! -e $BOOTDIR/grub/grub.conf ] && 
		   [ ! -e $BOOTDIR/grub/menu.lst ] && 
		   [ "`uname -m`"  != "ia64" ]; then
			if [ -e /sbin/lilo ]; then
				echo -e "Running LILO...."
				invoke_cmd "/sbin/lilo" "Updating lilo"
			fi
		fi
	fi
	return 0
}

# --------------------------------------------------------- #
# install_driver ()	                                    #
# Install the driver in /lib/modules. Backups old           #
# driver						    #
#							    #
# Parameter:	None					    # 
# Returns: 0 on success, 1 on fail			    #
# --------------------------------------------------------- #
function install_driver() {
	# Append kernel name to file containing list of installed on kernels
	echo $KERNEL_VER >> $INSTALL_LOG
	# Branch differently if the make succeeded or not
	MSG_DISPLAYED=0
	for MODULE in ${MODULE_NAME[@]} 
	do
		if [ ! -e ${MODULE}.o ]; then
			echo "Module $MODULE does not exist"
			#exit 1
		else
			#backup the old driver
			if [ $MSG_DISPLAYED = 0 ] && [ -e $SCSIDIR/${MODULE}.o ]; then
				echo -e "Creating backup of old driver in $QL_BACKUP...."
				echo -e ""
				MSG_DISPLAYED=1
			fi
			rename_driver "${MODULE}"
		fi
	done
	if [ -e $SCSIDIR ]; then
		echo "Installing driver in $SCSIDIR...."
		invoke_cmd " make -s install $SCSIDIR/"

		echo -e "Building module dependency...."
		# Module dependency
		invoke_cmd "/sbin/depmod -a" "depmod"
		ret_status=$?
		if [ "$ret_status" -ne "0" ]; then
			echo "depmod...failed"
			echo "Please restore backed up files from $QL_BACKUP"
			return $ret_status
		fi
#		depmod -a 
	else 
		echo -e "New driver could not be copied to $SCSIDIR, \
				          could not find proper directory" 
	fi

	# Make clean
	make clean >/dev/null
}

# --------------------------------------------------------- #
# detect_isp ()                                             #
# Detects if the installed HBA is of typed FC or iSCSI      #
# Fills up the global QL_HBA_LIST array with the installed  #
# HBAs                                                      #
# Parameter:                                                #
#       $1: Array of FC or iSCSI ISPs to be detected        #
# Returns: QL_SUCCESS if HBA found. QL_FAIL if no HBA found #
# --------------------------------------------------------- #
function detect_isp() {
        ISP_LIST=( `echo "$1"` )
        debug_print "in detect_isp: ISP_LIST"
        debug_print  ${ISP_LIST[@]}
        ret_status=1
        QL_HBA_LIST=() #Empty array
        # Get the list of line which has the keyword qlogic
        # Assumed that qlogic will be there in the output
        LSPCI=`lspci | grep -i qlogic`
        if [ $? = 0 ]; then
                # check if the installed HBA ISP is one in the
                # input array list
                for LSPCI_LINE in $LSPCI
                do
                        for ISP in "${ISP_LIST[@]}"
                        do
                                echo "$LSPCI_LINE" | grep $ISP >& /dev/null
                                if [ $? = 0 ]; then
                                        debug_print "Found an ISP = $ISP"
                                        # Store it in our array
                                        QL_HBA_LIST[$ISP]=$ISP;
                                        ret_status=0
                                fi
                        done
                done
        else
                # Could not find qlogic device
                # This could be because either `lspci` does
                # not work on this system or QLogic device is
                # not installed on this system.
                # Ask user to force for installation
                ret_status=1
        fi
        return $ret_status;
}
# --------------------------------------------------------- #
# restore_mod_conf()                                        #
# Restore any modules that installer removed from the       #
# modules configuration file during installation.           #
# Parameter:                                                #
#       $1      None                                        #
# Returns: None                                             #
# --------------------------------------------------------- #
restore_mod_conf() {
        #get the entries from  $QL_MOD_REMOVED
        # and put them back. Note this operation must
        # be performed after installer changes are removed
        # from modules conf. files

        if [ -f $QL_MOD_REMOVED ]; then
		if [ "$DISTRO" = "$SUSE" ]; then 
                        local MODULES=`cat $SYSCONFIG | grep INITRD_MODULES |
                        awk 'BEGIN {FS="="} {print $2}'| sed 's/\"//g'`

                        for MODULE in `cat $QL_MOD_REMOVED`
                        do
                                MODULES="$MODULES $MODULE"
                        done

                        sed "s/INITRD_MODULES.*/INITRD_MODULES=\"$MODULES\"/"  $SYSCONFIG >/tmp/kernel
                        mv -f /tmp/kernel $SYSCONFIG
                else #RedHat/AsianUx
                         cat $QL_MOD_REMOVED >> $MODCONF
                fi
        fi
}


function restore_original() {
        # Restore driver
	MSG_DISPLAYED=0
        for MODULE in ${MODULE_NAME[@]}
        do
                BACKUP_DRIVER=$QL_BACKUP/$MODULE.o-$KERNEL_VER.bak
                BACKUP_UPDATE_DRIVER=$SCSI_UPDATE_DIR/$MODULE-old.o
		local ORIGINAL_DRIVER=$SCSI_UPDATE_DIR/$MODULE.o
                if [ -e $BACKUP_DRIVER ]; then
			if [ $MSG_DISPLAYED = 0 ]; then
				echo -e "Restoring original QLogic drivers...."
				echo -e ""
				MSG_DISPLAYED=1
			fi
                        mv -f $BACKUP_DRIVER $SCSIDIR/$MODULE.o
               	else # No backup then remove the one in the /lib/modules/.... 
			rm -f $SCSIDIR/$MODULE.o
		fi

		# restore the update driver
		if [ -f $BACKUP_UPDATE_DRIVER ]; then
			mv -f $BACKUP_UPDATE_DRIVER $ORIGINAL_DRIVER
		fi
        done
        # Restore modules.conf
#        BACKUP_MOD=$QL_BACKUP/$MOD_FILE-$KERNEL_VER.bak
#        if [ -e $BACKUP_MOD ]; then
#		echo -e "Restoring original modules.conf...."
#		echo -e ""
#                cp -f $BACKUP_MOD $MODCONF
#        fi
                                                                                                 
        # Restore kernel For SuSE
        # Restore initrd
	#echo -e "Rebuilding ramdisk...."

	if [ -f $QL_CONF_ENTRIES ]; then
		grep -v -f $QL_CONF_ENTRIES $MODCONF > $QL_BACKUP/$MOD_FILE
		if [ -f $QL_BACKUP/$MOD_FILE ]; then
			echo -e "Removing installation changes from $MODCONF...."
			mv -f $QL_BACKUP/$MOD_FILE $MODCONF	
		else
			echo "Unable to update $MODCONF file"
			return;
		fi
	fi

        if [ "$DISTRO" = "$SUSE" ]; then
		#From the "INITRD_MODULES" line in the /etc/sysconfig/kernel 
		# remove entries present in $QL_KERNEL_ENTRIES
		if [ -f "$QL_KERNEL_ENTRIES" ]; then
			INITRD_LINE=`grep "INITRD_MODULES" $SYSCONFIG` 
			
			for MODULES in `cat $QL_KERNEL_ENTRIES`
			do
				INITRD_LINE=`echo $INITRD_LINE | sed "s/$MODULES//"`
			done 
			echo -e "Removing installation changes from $SYSCONFIG...."
			echo -e ""
			sed "s/INITRD_MODULES.*/$INITRD_LINE/"  $SYSCONFIG > /tmp/kernel
			cp -f /tmp/kernel $SYSCONFIG
		fi
        fi
               
	restore_mod_conf
	create_ramdisk "" $NO_BACKUP

        # All done, remove the contents of backup dir
#	if [ -e "$QL_SOURCE/backup" ]; then
#        	rm -rf $QL_SOURCE/backup
#	fi
}

