/*
 * Architecture specific (alpha) functions for Linux crash dumps.
 *
 * 2.3 kernel modifications by: Matt D. Robinson (yakker@turbolinux.com)
 * Copyright 2000 TurboLinux, Inc.  All rights reserved.
 * 2.4 modifications by Matt D. Robinson (yakker@aparity.com)
 * Copyright 2001 Matt D. Robinson.  All rights reserved.
 * 
 * This code is released under version 2 of the GNU GPL.
 */

/*
 * The hooks for dumping the kernel virtual memory to disk are in this
 * file.  Any time a modification is made to the virtual memory mechanism,
 * these routines must be changed to use the new mechanisms.
 */
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/dump.h>
#include <linux/mm.h>

/*
 * Name: __dump_configure_header()
 * Func: Configure the dump header with all proper values.
 */
int
__dump_configure_header(struct pt_regs *regs)
{
	dump_header_asm.dha_magic_number = DUMP_ASM_MAGIC_NUMBER;
	dump_header_asm.dha_version = DUMP_ASM_VERSION_NUMBER;
	dump_header_asm.dha_header_size = sizeof(dump_header_asm_t);

	/* copy the registers if they are valid */
	if (regs) {
		memcpy((void *)&(dump_header_asm.dha_regs),
			(const void *)regs, sizeof(struct pt_regs));
	}

	return (1);
}

/*
 * Name: __dump_write_header()
 * Func: Update the header information with all architecture specific
 *       information.
 */
int
__dump_write_header(char *dpage)
{
	memcpy((void *)(dpage + sizeof(dump_header_t)),
		(const void *)&dump_header_asm, sizeof(dump_header_asm_t));

	return (1);
}

/*
 * Name: __dump_init()
 * Func: Initialize the dumping routine process.  This is in case
 *       it's necessary in the future.
 */
void
__dump_init(uint64_t local_memory_start)
{
	/* return */
	return;
}

/*
 * Name: __dump_open()
 * Func: Open the dump device (architecture specific).  This is in
 *       case it's necessary in the future.
 */
void
__dump_open(struct file *dump_file, uint64_t memory_size)
{
	/* return */
	return;
}

/*
 * Name: __dump_silence_system()
 * Func: Do an architecture-specific silencing of the system.
 */
unsigned int
__dump_silence_system(unsigned int stage)
{
	/* return */
	return (0);
}

/*
 * Name: __dump_resume_system()
 * Func: Resume the system state in an architecture-specific way.
 */
unsigned int
__dump_resume_system(unsigned int stage)
{
	/* return */
	return (0);
}
