/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _DRIVERS_LKST_LKST_CORE_H_
#define _DRIVERS_LKST_LKST_CORE_H_

#include <linux/lkst_private.h>

#if DEBUG
/* enable messages for debug*/
#define LKST_DEBUG_MESSAGE
//TODO: call lkst_evhandlerprim_entry_log_static
#define LKST_DEBUG_LOG(ty, a1, a2, a3, a4) \
	({lkst_evhandlerprim_entry_log_static(ty,a1,a2,a3,a4);})

#else /* DEBUG */

#define LKST_DEBUG_LOG(ty, a1, a2, a3, a4)

#endif /* DEBUG */

#ifdef LKST_DEBUG_MESSAGE
#define LKST_DPRINTK(fmt, args...)  \
	printk(KERN_ERR "LKST(%s): " fmt, __FUNCTION__, ## args)

#define LKST_DPRINTK_IF(expr ,fmt, args...)  \
	({if (expr) {LKST_DPRINTK(fmt ,## args);}})
#else
#define LKST_DPRINTK(fmt, args...)
#define LKST_DPRINTK_IF(expr ,fmt, args...) 
#endif

#ifdef CONFIG_DEBUG_LKST_DEVICE_PER_CPU
#define LKST_NR_DEVS (LKST_NR_CPUS + 1)
#else
#define LKST_NR_DEVS (1)
#endif

/* intra-module symbols */
/* lkst_class.c */
extern int lkst_class_init(void);
extern void lkst_class_cleanup(void);
/* lkst_device.c */
extern int lkst_device_init(void);
extern void lkst_device_cleanup(void);
extern int lkst_buffer_read(struct lkst_log_buffer *lbuffer);
/* lkst_hook.c */
#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
extern int lkst_hook_proc_init(void);
extern void lkst_hook_proc_cleanup(void);
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/
extern void lkst_hook_init(void);
extern void lkst_hook_cleanup(void);
extern int  lkst_hook_set_exit(lkst_event_type etype, lkst_evhandler_id newid);
extern int lkst_hook_maskset_xchg(lkst_maskset_id mask_id_new,
				  lkst_maskset_id mask_id_old);
extern int lkst_hook_exit_update(lkst_evhandler_id evh_id);
extern struct lkst_etype_info * lkst_hook_etype_get(lkst_event_type etype);
extern void lkst_hook_etype_put(lkst_event_type etype);

/*
 *  The buffer read queue mechanism :
 * A buffer writing on is in a queue tail except for a spacial case that the 
 * buffer is read and not shifted-in, while reading it. When the reading 
 * process begin to read a buffer, the buffer should be dequeued. And 
 * read_buf field of lkst_current.owner point it. When the writing process 
 * begin to write (shift-in) a buffer, the buffer should be enqueued. 
 * This mechanism is adopted instead of the old wcnt/rcnt mechanism.
 * IOW, this is just a kind of LRU.(simple implementation:-) )
 */


/* if you give dequeued buffer, rqueue will be cleanuped.*/
static __inline__ \
int lkst_buffer_dequeue_inline(struct lkst_event_buffer * buffer) {
	/* ! Please check buffer != NULL !
	   ! and check buffer is queued !*/
	struct lkst_event_buffer *tmp;

	if ( buffer->rq_next == LKST_BUFFER_ID_VOID ) {  /* buffer is tail */
		lkst_current.rq_tail = buffer->rq_prev;
	} else {  /* next != NULL */
		if ( ! LKST_BUFFER_LOAD(buffer->rq_next, tmp)) 
			return -EBUSY; //TODO
		tmp->rq_prev = buffer->rq_prev;
	}

	if ( buffer->rq_prev == LKST_BUFFER_ID_VOID ) { /* buffer is head */
		lkst_current.rq_head = buffer->rq_next;
	} else { /* prev != NULL */
		if ( ! LKST_BUFFER_LOAD(buffer->rq_prev, tmp)) 
			LKST_ENCOUNTER_A_BUG; //TODO
		tmp->rq_next = buffer->rq_next;
	}
	buffer->rq_next = buffer->rq_prev = LKST_BUFFER_ID_VOID;
	return 0;
}

/*
 * This enqueue a buffer in read queue.
 */
static __inline__ \
int lkst_buffer_enqueue_inline(struct lkst_event_buffer * buffer) {
	/* ! Please check buffer != NULL ! 
	   ! and check buffer is dequeued !*/
	if ( lkst_current.rq_tail != buffer->id ) { 
		if ( buffer->rq_next == LKST_BUFFER_ID_VOID) {
			if ( lkst_current.rq_tail == LKST_BUFFER_ID_VOID ) { 
				/* then rqueue is empty.*/
				lkst_current.rq_tail = buffer->id;
				lkst_current.rq_head = buffer->id;
			} else {
				struct lkst_event_buffer * tail;
				if ( ! LKST_BUFFER_LOAD(lkst_current.rq_tail,
							tail) ) return -EBUSY;
				tail->rq_next = buffer->id;
				buffer->rq_prev = lkst_current.rq_tail;
				lkst_current.rq_tail = buffer->id;
			}
		} else {
			struct lkst_event_buffer *tmp, *tmp2, *tmp3;
			/*dequeue*/
			if (!LKST_BUFFER_LOAD(buffer->rq_next, tmp) ||
			    !LKST_BUFFER_LOAD(lkst_current.rq_tail, tmp3)) /* if tail is null, buffer is dequeued. No, buffer is queued.*/
				return -EBUSY;
			if ( buffer->rq_prev != LKST_BUFFER_ID_VOID ) {
				if (!LKST_BUFFER_LOAD(buffer->rq_prev, tmp2))
					return -EBUSY;
				tmp2->rq_next = buffer->rq_next;
			} else {
				lkst_current.rq_head = buffer->rq_next;
			}
			tmp->rq_prev = buffer->rq_prev;
			/*enqueue*/
			tmp3->rq_next = buffer->id;
			buffer->rq_prev = lkst_current.rq_tail;
			lkst_current.rq_tail = buffer->id;
			buffer->rq_next = LKST_BUFFER_ID_VOID;
		}
	}
	return 0;
}

#define EINVETYPE 0x10
#define EINVEVHND 0x11

#endif
