/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/stat.h>

#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include <linux/lkst_hook.h>
#include <linux/lkst_private.h>
#include "lkst_core.h"

// TODO: lock, please...

/* inter-module symbols */
EXPORT_SYMBOL(lkst_encounter_a_BUG);
EXPORT_SYMBOL(lkst_hook_etype_register);
EXPORT_SYMBOL(lkst_hook_etype_unregister);

/* define event type informations 
 (In other words, define lkst_einfo_##mnemonic ) */

#define LKST_ETYPE_DEF(event_type, hooktype, mnemonic, event_name, arg_1, arg_2, arg_3, arg_4) \
        LKST_EINFO_DEF_##hooktype(LKST_ET_SYM(mnemonic), mnemonic, event_name, arg_1, arg_2, arg_3, arg_4)

#define LKST_EINFO_DEF(etype, eflags, mnemonic, name,arg1,arg2,arg3,arg4, h) \
	extern struct hook h; \
	static LKST_ETYPE_INFO_DEF(etype, eflags, mnemonic, name,arg1,arg2,arg3,arg4,&h);

#define LKST_EINFO_DEF_NORMAL(etype, mnemonic, name,arg1,arg2,arg3,arg4) \
	LKST_EINFO_DEF(etype, LKST_EFG_NORMAL, mnemonic, name,arg1,arg2,arg3,arg4,LKST_HK_SYM(etype))
#define LKST_EINFO_DEF_INLINE(etype, mnemonic, name,arg1,arg2,arg3,arg4) \
	LKST_EINFO_DEF(etype, LKST_EFG_INLINE, mnemonic, name,arg1,arg2,arg3,arg4,LKST_HK_SYM(etype))
#define LKST_EINFO_DEF_MODULE(etype, mnemonic, name,arg1,arg2,arg3,arg4) \
	LKST_EINFO_DEF(etype, LKST_EFG_INLINE, mnemonic, name,arg1,arg2,arg3,arg4,LKST_HK_SYM(etype))
#define LKST_EINFO_DEF_UNDEF(etype, mnemonic, name,arg1,arg2,arg3,arg4) \
	static LKST_ETYPE_INFO_DEF(etype, LKST_EFG_UNDEF, mnemonic, name,arg1,arg2,arg3,arg4,NULL);

#include <linux/lkst_etype.h>

#undef  LKST_ETYPE_DEF

/* define an event type information table */
#define LKST_ETYPE_DEF(event_type, hooktype, mnemonic, event_name, arg_1, arg_2, arg_3, arg_4) \
	[event_type] = &LKST_ETYPE_INFO(mnemonic),

static struct lkst_etype_info *lkst_einfo_table[LKST_MASKSET_TABLE_LEN_MAX] = {
#include <linux/lkst_etype.h>
};
#undef  LKST_ETYPE_DEF


struct lkst_etype_info * lkst_hook_etype_get(lkst_event_type etype)
{
	return lkst_einfo_table[etype];
}

void lkst_hook_etype_put(lkst_event_type etype)
{;}

/* for pointer verification, etc... */
void * lkst_encounter_a_BUG (const char *file, int line)
{
	int etype;
	/* make all hook point disarmed*/
	for ( etype = 0; etype <= LKST_ETYPE_MAX ; etype++) {
		if(lkst_einfo_table[etype])
			hook_exit_disarm(&lkst_einfo_table[etype]->hook_rec);
	}
	/* Now we encounter a bug. and cannot help ourselves. 
	 Only what we can do is panic dump.*/
	panic("LKST encounter a BUG at %s : %d\n", file, line);
	return NULL;
}

int lkst_hook_set_exit(lkst_event_type etype, lkst_evhandler_id newid)
{
	int ret = 0;
	struct hook_rec* hook_rec;
	lkst_evhandler_func new_evhandler;

	if (lkst_einfo_table[etype] == NULL || 
	    lkst_einfo_table[etype]->hook == NULL || 
	    !(lkst_einfo_table[etype]->flags & LKST_EF_MASKABLE) ) 
		return -EINVETYPE;
	hook_rec = &lkst_einfo_table[etype]->hook_rec;

	/* unregister old exit (if exist)*/
	if (hook_rec->hook_head != NULL) {
		hook_exit_deregister(hook_rec);
		memset(hook_rec, 0, sizeof(struct hook_rec));
	}
	if (newid == LKST_EVHANDLER_ID_VOID) return 0;

	/* register new exit (if specified)*/
	if (! (new_evhandler = lkst.evhandler_table[newid].evhandler) ){
		return -EINVEVHND;
	}
	
	hook_rec->hook_exit = new_evhandler;
	ret = hook_exit_register(lkst_einfo_table[etype]->hook,
				 hook_rec);
	if ( ret == 0 ) {
		hook_exit_arm(hook_rec);
	}
	return ret;
}

/* 
 * Exchange maskset
 *  Called from lkst_trc_init_stage0, lkst_evhandlerprim_maskset_xchg_inline,
 *  lkst_trc_start_core and lkst_trc_stop_core.
 */
int lkst_hook_maskset_xchg(lkst_maskset_id mask_id_new,
			   lkst_maskset_id mask_id_old)
{
	int i;
	lkst_maskset *maskset_old;
	lkst_maskset *maskset_new;
	lkst_evhandler_id evh_id_old;
	lkst_evhandler_id evh_id_new;

	if (mask_id_old > LKST_MASKSET_ID_MAX) {
		mask_id_old = LKST_MASKSET_ID_RNOTHING;
	}
	if (mask_id_new > LKST_MASKSET_ID_MAX) {
		mask_id_new = LKST_MASKSET_ID_RDEFAULT;
	}

	if(mask_id_old == mask_id_new) return 0;

	if ( ! (maskset_old = lkst.maskset_table[mask_id_old].masksetp) || 
	     ! (maskset_new = lkst.maskset_table[mask_id_new].masksetp) )
		return -EINVAL;

	for (i = 0; i < LKST_MASKSET_TABLE_LEN_MAX; i++) {
		if (lkst_einfo_table[i] == NULL || 
		    lkst_einfo_table[i]->hook == NULL || 
		    !(lkst_einfo_table[i]->flags & LKST_EF_MASKABLE) ) 
			continue;
		evh_id_old = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset_old, i);
		evh_id_new = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset_new, i);

		if(evh_id_old != evh_id_new) {
			int ret;
			if ((ret = lkst_hook_set_exit(i, evh_id_new)) < 0 ){
				printk(KERN_ERR "LKST error(hook_maskset_xchg):"
				       "lkst_hook_set_exit failed: %d\n", ret);
				printk(KERN_ERR "Etype: %d, New handler ID: %d, Old Handler ID: %d\n",
				       i, evh_id_new, evh_id_old);
				return -EINVAL;
			}
		}
	}
	return 0;
}
/* 
 * Update hook exit
 *  Called from lkst_evhandler_register/lkst_evhandler_unregister
 */
int lkst_hook_exit_update(lkst_evhandler_id evh_id)
{
	lkst_maskset *maskset = NULL;
	int i, ret = 0;

	if ( evh_id > LKST_EVHANDLER_ID_MAX ) return -EINVAL;
	if ( ! (maskset = lkst.current_masksetp ) )
		return -EBUSY;/*Cause, current_masksetp is never set NULL*/

	ret = 0;
	for(i=0; i<LKST_MASKSET_TABLE_LEN_MAX; i++){
		if (evh_id == LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, i)) {
			if ( lkst_hook_set_exit(i, evh_id) ) {
				printk(KERN_ERR "LKST error(hook_exit_update):"
				       " lkst_hook_set_exit(%d, %d) failed\n",
				       i , evh_id);
				ret = -EINVAL;
			}
		}
	}
	return ret;
}

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS

static unsigned long lkst_etype_info_len(struct lkst_etype_info *einfo)
{
	unsigned long size = 0;
	int i;
	if (einfo == NULL )return 0;
	size += 6;//0x000,
	size += strlen(einfo->mnemonic)+1;//mnemonic,
	size += 7;//0x0000,
	if(einfo->name)
		size += strlen(einfo->name)+3;//"name",
	else
		size += 3;
	for(i=0; i<4; i++) {
		if(einfo->args[i])
			size += strlen(einfo->args[i])+3;//"arg", o "arg"\n
		else
			size += 3;
	}
	return size;
}

/* iterator */
static inline int get_next_etype_index(int i)
{
	if (i > LKST_ETYPE_MAX || i < -1) return LKST_ETYPE_VOID;
	for ( i+=1 ;i<LKST_ETYPE_MAX; i++) {
		if( lkst_einfo_table[i] != NULL) break;
	}
	return i;
}

static void *lkst_hook_etype_seq_start(struct seq_file *m, loff_t *pos)
{
	loff_t n = *pos;
	lkst_event_type i;
	
	for(i = get_next_etype_index(-1);
	    i != LKST_ETYPE_VOID && n != 0; 
	    i = get_next_etype_index(i), n--);
	
	if (i == LKST_ETYPE_VOID) return NULL;
	return lkst_einfo_table[i];
}

static void *lkst_hook_etype_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	struct lkst_etype_info *einfo = v;
	lkst_event_type i;

	(*pos)++;
	i = get_next_etype_index(einfo->etype);

	if (i == LKST_ETYPE_VOID) return NULL;
	return lkst_einfo_table[i];
}

static void lkst_hook_etype_seq_stop(struct seq_file *m, void *v)
{
/*	if (v) {
		struct lkst_etype_info *einfo = v;
		lkst_hook_etype_put(einfo->etype);
	}*/
}

static int lkst_hook_etype_seq_show(struct seq_file *m, void *v)
{
	struct lkst_etype_info *einfo = v;
	int i;

	if (einfo == NULL) return 0;
	seq_printf(m, "0x%03x,%s,0x%04x,\"%s\"",
		   einfo->etype, einfo->mnemonic,
		   einfo->flags & LKST_EF_MASK,
		   einfo->name?einfo->name:"");
	for(i=0;i<4;i++)
		seq_printf(m,",\"%s\"",
			   einfo->args[i]?einfo->args[i]:"");
	seq_putc(m, '\n');
	return 0;
}

static struct seq_operations lkst_hook_etype_op = {
	.start	= lkst_hook_etype_seq_start,
	.next	= lkst_hook_etype_seq_next,
	.stop	= lkst_hook_etype_seq_stop,
	.show	= lkst_hook_etype_seq_show
};

static int lkst_hook_etype_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &lkst_hook_etype_op);
}

static struct file_operations proc_lkst_etypes_operations = {
	.owner		= THIS_MODULE,
	.open		= lkst_hook_etype_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};
static struct proc_dir_entry *lkst_proc_ent = NULL;
static unsigned long lkst_proc_ent_size = 0;
//struct proc_dir_entry *proc_lkst_etypes_dir;

int lkst_hook_proc_init(void)
{
	lkst_proc_ent = create_proc_entry("lkst_etypes", S_IRUGO,
					  &proc_root);
	if (!lkst_proc_ent)
		return -ENOMEM;
	lkst_proc_ent->proc_fops = &proc_lkst_etypes_operations;
	lkst_proc_ent->size = lkst_proc_ent_size;
	LKST_DPRINTK("proc interface created\n");
	return 0;
}

void lkst_hook_proc_cleanup(void)
{
	lkst_proc_ent = NULL;
	remove_proc_entry("lkst_etypes", &proc_root);
}

#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/


int lkst_hook_etype_register(struct lkst_etype_info *info)
{
	int etype;
	lkst_evhandler_id evhandler_id;
	lkst_maskset *maskset = NULL;

	if (info == NULL || info->hook == NULL ) return -EINVAL;
	etype = info->etype;
	if (LKST_ETYPE_RESERVED(etype)) return -EINVAL;
	if (lkst_einfo_table[etype]) return -EBUSY;
	lkst_einfo_table[etype] = info;

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	lkst_proc_ent_size += lkst_etype_info_len(info);
	if(lkst_proc_ent) lkst_proc_ent->size = lkst_proc_ent_size;
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/

	if ( ! (maskset = lkst.current_masksetp) ) {
		LKST_DPRINTK("currnet maskset temprary busy, or broken.\n");
		return -EBUSY;
	}
	evhandler_id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, etype);
	lkst_hook_set_exit(etype, evhandler_id);
	return etype;
}

int lkst_hook_etype_unregister(struct lkst_etype_info *info)
{
	int etype;
	if (info == NULL) return -EINVAL;

	etype = info->etype;
	if (LKST_ETYPE_RESERVED(etype) || 
	    lkst_einfo_table[etype] != info) return -EBUSY;

	if (info->hook_rec.hook_head != NULL)
		hook_exit_deregister(&(info->hook_rec));

	lkst_einfo_table[etype] = NULL;

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	lkst_proc_ent_size -= lkst_etype_info_len(info);
	if(lkst_proc_ent) lkst_proc_ent->size = lkst_proc_ent_size;
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/
	return etype;
}

void lkst_hook_init(void)
{
#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	int i,j;
	lkst_proc_ent_size = 0;
	for(i = get_next_etype_index(-1), j=0;
	    i != LKST_ETYPE_VOID;
	    i = get_next_etype_index(i), j++) {
		lkst_proc_ent_size += lkst_etype_info_len(lkst_einfo_table[i]);
	}
	LKST_DPRINTK("All embeded event type counted : num = %d, size = %lu\n",j,lkst_proc_ent_size);
#endif /*CONFIG_DEBUG_LKST_ETYPES_PROCFS*/
}

void lkst_hook_cleanup(void)
{
	int i;
	/* Cleanup of hook */
	for (i=0;i<LKST_ETYPE_MAX; i++) {
		/*deregister hooks*/
		if ( lkst_einfo_table[i] ) {
			if (lkst_einfo_table[i]->hook_rec.hook_head != NULL)
				hook_exit_deregister(&lkst_einfo_table[i]->hook_rec);
		}
	}
}

