/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by T.Nakamura <tetsu@crl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _ARCH_I386_LKST_PRIVATE_H
#define _ARCH_I386_LKST_PRIVATE_H

#include <asm/lkst.h>

#define LKST_AVAILABLE_MEMORY_MIN	(LKST_BUFFER_SIZE_MIN * 256)
#define LKST_AVAILABLE_MEMORY_MAX	(1024*1024*1024)	/* dummy */

#ifdef __KERNEL__
#include <asm/msr.h>
#include <asm/atomic.h>
#include <asm/timex.h>		/* for cpu_khz */

/* format strings for printk() */
#define LKST_FMT_TSC	"%016llu"	/* lkst_tsc_t */
#define LKST_FMT_SZT	"%u"		/* size_t */

/**
 * atomic_read_and_inc - read and increment atomic variable
 * @v: pointer of type atomic_t
 * 
 * Atomically increments @v by 1.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */

static __inline__ int atomic_read_and_add(int i, atomic_t * v)
{
	__asm__ __volatile__(LOCK "xadd %1,%0":"+m"(v->counter), "=r"(i)
			     :"0"(v->counter), "1"(i):"memory");
	return (i);
}

/**
 * local_atomic_read_and_inc - read and increment a variable on local processor
 * @i: integer value to add
 * @v: pointer of type int. 
 */
static __inline__ int local_atomic_read_and_add(int i, int *v)
{
	__asm__ __volatile__("xadd %1,%0":"+m"(*v), "=r"(i)
			     :"0"(*v), "1"(i):"memory");
	return (i);
}

/*
 * function lkst_evhandlerprim_mc()
 * A primitive part of an Event-handler. This function returns current
 * value of machine-cycle-counter.
 * 
 *  Arguments
 *    NONE
 *
 *  returns
 *    tsc:	current value of machine-cycle-counter.
 * 
 */
static __inline__ lkst_tsc_t lkst_evhandlerprim_mc(void)
{
	register lkst_tsc_t tsc;

	/* Read the Time Stamp Counter */
	rdtscll(tsc);
	return tsc;
}


/*
 * get variable "cpu_khz".
 */
#define	LKST_GET_CPU_KHZ() ((lkst_cpu_freq_t)cpu_khz)

#define LKST_LIMIT_ADDRESS_CHECK(buffer, size, limit)			\
	( (lkst_ptr_t)((char *) buffer + size) <= (lkst_ptr_t)limit)

#endif /*__KERNEL__*/

#endif /* _ARCH_I386_LKST_PRIVATE_H */
