/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

/* IA64 architecture dependant events */
/* Do not check double (or more) including.
 This file will be included many times.*/

//0x004 event
LKST_ETYPE_DEF(0x004, NORMAL, PROCESS_LTHREADGEN, "kernel_thread", \
	"pointer to kernel thread fn ", \
	"argument of kernel thread arg ", \
	"flag", \
	NULL)

//0x010 event
LKST_ETYPE_DEF(0x010, NORMAL, INT_HARDWARE_ENTRY, "interrupt_hardware_entry", \
	"IRQ number irq ", \
	"interrupt status status ", \
	"pointer to register stack", \
	NULL)

//0x030 event
LKST_ETYPE_DEF(0x030, NORMAL, SYSCALL_ENTRY, "system_call_entry", \
	"handler address", \
	"call number", \
	NULL, \
	NULL)

//0x031 event
LKST_ETYPE_DEF(0x031, NORMAL, SYSCALL_EXIT, "system_call_exit", \
	"handler address", \
	"return value errno ", \
	NULL, \
	NULL)

//0x020 event
LKST_ETYPE_DEF(0x020, NORMAL, EXCEPT_PGFLT_ENTRY, "except_pgflt_entry", \
	"fault address ifa ", \
	"isr", \
	"ipsr", \
	"iip")

//0x021 event
LKST_ETYPE_DEF(0x021, NORMAL, EXCEPT_PGFLT_EXIT, "except_pgflt_exit", \
	"fault address ifa ", \
	"isr", \
	"ipsr", \
	"iip")

//0x022 event
LKST_ETYPE_DEF(0x022, NORMAL, EXCEPT_ILLOP_ENTRY, "except_illop_entry", \
	"ec", \
	NULL, \
	"ipsr", \
	"iip")

//0x023 event
LKST_ETYPE_DEF(0x023, NORMAL, EXCEPT_ILLOP_EXIT, "except_illop_exit", \
	"ec", \
	NULL, \
	"ipsr", \
	"iip")

//0x024 event
LKST_ETYPE_DEF(0x024, NORMAL, EXCEPT_BADBRK_ENTRY, "except_badbrk_entry", \
	"break number iim ", \
	NULL, \
	"ipsr", \
	"iip")

//0x025 event
LKST_ETYPE_DEF(0x025, NORMAL, EXCEPT_BADBRK_EXIT, "except_badbrk_exit", \
	"break number iim ", \
	NULL, \
	"ipsr", \
	"iip")

//0x026 event
LKST_ETYPE_DEF(0x026, NORMAL, EXCEPT_FAULT_ENTRY, "except_fault_entry", \
	"fault vector number", \
	"isr", \
	"ipsr", \
	"iip")

//0x027 event
LKST_ETYPE_DEF(0x027, NORMAL, EXCEPT_FAULT_EXIT, "except_fault_exit", \
	"fault vector number", \
	"isr", \
	"ipsr", \
	"iip")

//0x028 event
LKST_ETYPE_DEF(0x028, NORMAL, EXCEPT_UNALIGN_ENTRY, "except_unalign_entry", \
	"ifa", \
	NULL, \
	"ipsr", \
	"iip")

//0x029 event
LKST_ETYPE_DEF(0x029, NORMAL, EXCEPT_UNALIGN_EXIT, "except_unalign_exit", \
	"ifa", \
	NULL, \
	"ipsr", \
	"iip")

/* Events concerned with spinlocks should have event_type as 0x08x. */
//0x080 event
LKST_ETYPE_DEF(0x080, INLINE, LK_SPINLOCK, "spin_lock", \
	"call address", \
	"lock", \
	"flags", \
	NULL)

//0x081 event
LKST_ETYPE_DEF(0x081, INLINE, LK_SPINTRYLOCK, "spin_trylock", \
	"call address", \
	"lock", \
	NULL, \
	NULL)

//0x082 event
LKST_ETYPE_DEF(0x082, INLINE, LK_SPINUNLOCK, "spin_unlock", \
	"call address", \
	"lock", \
	NULL, \
	NULL)

//0x083 event
LKST_ETYPE_DEF(0x083, INLINE, LK_WRLOCK, "write_lock", \
	"call address", \
	"rwlock", \
	NULL, \
	NULL)

//0x084 event
LKST_ETYPE_DEF(0x084, INLINE, LK_WRTRYLOCK, "write_trylock", \
	"call address", \
	"rwlock", \
	NULL, \
	NULL)

//0x085 event
LKST_ETYPE_DEF(0x085, INLINE, LK_WRUNLOCK, "write_unlock", \
	"call address", \
	"rwlock", \
	NULL, \
	NULL)

//0x086 event
LKST_ETYPE_DEF(0x086, INLINE, LK_RDLOCK, "read_lock", \
	"call address", \
	"rwlock", \
	NULL, \
	NULL)

//0x087 event
LKST_ETYPE_DEF(0x087, INLINE, LK_RDUNLOCK, "read_unlock", \
	"call address", \
	"rwlock", \
	NULL, \
	NULL)

//0x090 event
LKST_ETYPE_DEF(0x090, INLINE, O_PORTIN, "portin", 
	"port address/bite-width", \
	"value", \
	"call address", \
	NULL)

//0x091 event
LKST_ETYPE_DEF(0x091, INLINE, O_PORTOUT, "portout", \
	"port address/bite-width", \
	"value", \
	"call address", \
	NULL)

//0x0b0 event
LKST_ETYPE_DEF(0x0b0, NORMAL, OOPS_PGFAULT, "page_fault_oops", \
	"accessed virtual address", \
	"fault code address", \
	"error code", \
	NULL)

