/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _ARCH_IA64_LKST_PRIVATE_H
#define _ARCH_IA64_LKST_PRIVATE_H

#include <asm/lkst.h>

#define LKST_AVAILABLE_MEMORY_MIN	(LKST_BUFFER_SIZE_MIN * 256)
#define LKST_AVAILABLE_MEMORY_MAX	(64UL*1024*1024*1024)	/* 64GB */

#ifdef __KERNEL__
#include <asm/atomic.h>
#include <asm/timex.h>		/* for get_cycles() */

/* format strings for printk() */
#define LKST_FMT_TSC	"%016lu"	/* lkst_tsc_t */
#define LKST_FMT_SZT	"%lu"		/* size_t */

/**
 * atomic_read_and_add - add integer to atomic variable
 * @i: integer value to add
 * @v: pointer of type atomic_t
 * 
 * Atomically adds @i to @v.  Note that the guaranteed
 * useful range of an atomic_t is only 24 bits.
 */

static __inline__ int
atomic_read_and_add (int i, atomic_t *v)
{
	__s32 old, new;
	CMPXCHG_BUGCHECK_DECL

	do {
		CMPXCHG_BUGCHECK(v);
		old = atomic_read(v);
		new = old + i;
	} while (ia64_cmpxchg("acq", v, old, old + i, sizeof(atomic_t)) != old);

	return old;
}

/**
 * local_atomic_read_and_inc - read and increment a variable on local processor
 * @i: integer value to add
 * @v: pointer of type int. 
 */
static __inline__ int local_atomic_read_and_add(int i, int *v)
{
	int old, new;
	CMPXCHG_BUGCHECK_DECL

	do {
		CMPXCHG_BUGCHECK(v);
		old = *v;
		new = old + i;
	} while (ia64_cmpxchg("acq", v, old, old + i, sizeof(int)) != old);

	return old;
}

/*
 * function lkst_evhandlerprim_mc()
 * A primitive part of an Event-handler. This function returns current
 * value of machine-cycle-counter.
 * 
 *  Arguments
 *    NONE
 *
 *  returns
 *    tsc:	current value of machine-cycle-counter.
 * 
 */
static __inline__ u_int64_t lkst_evhandlerprim_mc(void)
{
	__u64 itc;

	/* Read the Interval Timer Counter */
	__asm__ __volatile__ ("mov %0=ar.itc" : "=r"(itc));

	return itc;
}


/*
 * get variable "cpu's khz".
 */
#define	LKST_GET_CPU_KHZ() ((lkst_cpu_freq_t)(local_cpu_data->itc_freq / 1000))

#define LKST_LIMIT_ADDRESS_CHECK(buffer, size, limit)			\
	(((void *) ((char *) buffer + size)) <= limit)

#endif /*__KERNEL__*/

#endif /* _ARCH_IA64_LKST_PRIVATE_H */
