#ifndef __ASM_PPC_HOOK_PRIVATE_H
#define __ASM_PPC_HOOK_PRIVATE_H
/*
 * Kernel Hooks optimized for PPC.
 * 
 * Authors: Mike Grundy <grundym@us.ibm.com> s390x
 */
#include <asm-generic/hook_private.h>
/*
 * Sanity check the hook location for valid instructions at hook location.
 * At hook location, we should find these instructions:
 *	38 00 00 00       	li    	r0,0
 *	2c 00 00 00            	cmpwi	r0,0
 *	
 * We can check for li and cmpwi instructions. As these instructions encode
 * the register name in the second byte and the register cannot be predicted, 
 * we mask out the bits corresponding to registers in the opcode before comparing.
 * PPC opcodes are six bits, hence mask of 0xFC
 */
#define OPCODE_MOV1			0x38 /* LI (really an extended mnemonic for addi */   
#define OPCODE_MOV1_MASK		0xFC
/* Compiler generates 2c 00 00 00     cmpwi   r0,0 */
		
static inline int is_asm_hook(unsigned char * addr)
{
	if (!addr)
		return 0;
	
	if((addr[0] & OPCODE_MOV1_MASK) == OPCODE_MOV1) {
		if (*((unsigned short *)(addr+1)) == 0)
			return 1;
	}
	return 0;
}
#endif /* __ASM_PPC_HOOK_PRIVATE_H */
