/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.SERIZAWA <serizawa@sdl.hitachi.co.jp>
 * Updated by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _LINUX_LKST_H
#define _LINUX_LKST_H

#ifdef __KERNEL__
# include <linux/config.h>
# include <linux/types.h>
#endif				/* __KERNEL__ */
#include <linux/threads.h>
#include <linux/ioctl.h>

#include <linux/lkst_hook.h>
#include <asm/lkst.h>

/* LKST version */
#define LKST_VERSION_MAJOR	2
#define LKST_VERSION_MINOR	2
#define LKST_VERSION_REVISION	1
#define LKST_VERSION_STR	"2.2.1"

/* Max number of CPUs */
#define LKST_CPU_MAX	NR_CPUS /*obsolate*/
#define LKST_NR_CPUS	NR_CPUS

/* ioctl requests */
#define LKST_IOC_MAGIC			'l'
#define LKST_IOC_TRC_STATUS		_IO(LKST_IOC_MAGIC, 1)
#define LKST_IOC_TRC_START		_IO(LKST_IOC_MAGIC, 2)
#define LKST_IOC_TRC_STOP		_IO(LKST_IOC_MAGIC, 3)

#define LKST_IOC_MASKSET_READ		_IO(LKST_IOC_MAGIC, 4)
#define LKST_IOC_MASKSET_WRITE		_IO(LKST_IOC_MAGIC, 5)
#define LKST_IOC_MASKSET_SET		_IO(LKST_IOC_MAGIC, 6)
#define LKST_IOC_MASKSET_LIST		_IO(LKST_IOC_MAGIC, 7)
#define LKST_IOC_MASKSET_DELETE		_IO(LKST_IOC_MAGIC, 8)

#define LKST_IOC_EVHANDLER_LIST		_IO(LKST_IOC_MAGIC, 9)
#define LKST_IOC_EVHANDLER_CTRL		_IO(LKST_IOC_MAGIC, 10)

#define LKST_IOC_BUFFER_READ		_IO(LKST_IOC_MAGIC, 11)
#define LKST_IOC_BUFFER_CREATE		_IO(LKST_IOC_MAGIC, 12)
#define LKST_IOC_BUFFER_SHIFT		_IO(LKST_IOC_MAGIC, 13)
#define LKST_IOC_BUFFER_JUMP		_IO(LKST_IOC_MAGIC, 14)
#define LKST_IOC_BUFFER_DELETE		_IO(LKST_IOC_MAGIC, 15)
#define LKST_IOC_BUFFER_LIST		_IO(LKST_IOC_MAGIC, 16)
#define LKST_IOC_BUFFER_SETRMOD		_IO(LKST_IOC_MAGIC, 17)
#define LKST_IOC_BUFFER_LINK		_IO(LKST_IOC_MAGIC, 18)

#define LKST_IOC_ENTRY_LOG		_IO(LKST_IOC_MAGIC, 19)

#define LKST_IOC_MASKSET_GETID		_IO(LKST_IOC_MAGIC, 20)
#define LKST_IOC_EVHANDLER_GETID	_IO(LKST_IOC_MAGIC, 21)

/* maskset & buffer */
typedef u_int8_t lkst_maskset_id;
typedef u_int8_t lkst_buffer_id;
typedef u_int8_t lkst_evhandler_id;

/* Maskset to record no event */
#define LKST_MASKSET_ID_RNOTHING	0
/* Maskset to record all events */
#define LKST_MASKSET_ID_RALL		1
/* Maskset to record events specified when kernel compilation */
#define LKST_MASKSET_ID_RDEFAULT	2

/* LKST_MASKSET_TBL_MAX means size of maskset_table. (cf. LKST_MASKSET_TBL_LEN_MAX */
#define LKST_MASKSET_TBL_MAX	((lkst_maskset_id) 0xFF)
#define LKST_MASKSET_ID_MAX	((lkst_maskset_id) (LKST_MASKSET_TBL_MAX - 1))
#define LKST_MASKSET_ID_VOID	((lkst_maskset_id) (LKST_MASKSET_ID_MAX + 1))

#define LKST_EVHANDLER_TBL_MAX			((lkst_evhandler_id) 0xFF)
#define LKST_EVHANDLER_ID_MAX						\
	((lkst_evhandler_id) (LKST_EVHANDLER_TBL_MAX - 1))
#define LKST_EVHANDLER_ID_VOID						\
	((lkst_evhandler_id) (LKST_EVHANDLER_ID_MAX + 1))

/* Definitions of event-handler ID.
 *	(obsolate)EVHANDLER_ID_DMX:		for Dynamic Maskset eXchange
 *	EVHANDLER_ID_DEFAULT:		default event-handler
 *	EVHANDLER_ID_BUFFER_SHIFT:	shift buffers on buffer-overflow
 *	(obsolate)EVHANDLER_ID_BUFFER_SHIFT_DW:	shift buffers and Wakeup LKST Daemon
 *	(optional)EVHANDLER_ID_EXCEPTION:		for debug of exception loop
 *	(obsolate)EVHANDLER_ID_MASKSET_XCHG:	do dynamic maskset exchange
 *	EVHANDLER_ID_LOG_STATIC:	logging to static buffer
 */
//#define LKST_EVHANDLER_ID_DMX 			((lkst_evhandler_id) 0x00)
#define LKST_EVHANDLER_ID_DEFAULT		((lkst_evhandler_id) 0x01)
#define LKST_EVHANDLER_ID_BUFFER_SHIFT		((lkst_evhandler_id) 0x02)
//#define LKST_EVHANDLER_ID_BUFFER_SHIFT_DW	((lkst_evhandler_id) 0x03)
#define LKST_EVHANDLER_ID_EXCEPTION		((lkst_evhandler_id) 0x04)
//#define LKST_EVHANDLER_ID_MASKSET_XCHG		((lkst_evhandler_id) 0x05)
#define LKST_EVHANDLER_ID_LOG_STATIC		((lkst_evhandler_id) 0x06)

#define LKST_EVHANDLER_ID_RESERVED		((lkst_evhandler_id) 0x1f)

#define LKST_BUFFER_TBL_MAX	((lkst_buffer_id) 0xFF)
#define LKST_BUFFER_ID_MAX	((lkst_buffer_id) (LKST_BUFFER_TBL_MAX - 1))
#define LKST_BUFFER_ID_VOID	((lkst_buffer_id) (LKST_BUFFER_ID_MAX + 1))


struct lkst_version {
	int major;
	int minor;
	int revision;
};

struct lkst_status_param {
	int	nr_cpus;	/* configurated number of CPUs (= NR_CPUS) */
	int	maskset_num;	/* total number of registered masksets */
	int	evhandler_num;	/* total number of registered event handlers */
	int	static_buffer_recid;	/* recid of the static buffer*/
	size_t	static_buffer_size;	/* size of the static buffer*/
	struct lkst_version version;	/* lkst version */
	lkst_maskset_id	current_maskset_id;
				/* current selected maskset ID */
	struct {
		lkst_buffer_id	write;	/* current writing buffer ID */
		lkst_buffer_id	read;	/* current reading buffer ID */
	} buffer_id[LKST_NR_CPUS_MAX];
};

struct lkst_entry_args {
	int event_type;		/* corresponding event type */	
	lkst_arg_t arg1;	/* 1st argument acquired at a trace point */
	lkst_arg_t arg2;	/* 2nd argument acquired at a trace point */
	lkst_arg_t arg3;	/* 3rd argument acquired at a trace point */
	lkst_arg_t arg4;	/* 4th argument acquired at a trace point */
};


#ifdef __KERNEL__

extern int lkst_entry_log(int event_type, lkst_arg_t arg1, lkst_arg_t arg2,
			   lkst_arg_t arg3, lkst_arg_t arg4);

extern void lkst_trc_init_stage0(void);

extern int lkst_trc_init_stage1(void);

extern int lkst_trc_init_main(void);

extern void lkst_time_set(int from_settimeofday);

extern int lkst_trc_status(struct lkst_status_param *param);

extern int lkst_trc_start(void);

extern int lkst_trc_stop(void);

#endif /* __KERNEL__ */

#endif				/* _LINUX_LKST_H */
