/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/
#ifndef _LINUX_LKST_CLASS_H
#define _LINUX_LKST_CLASS_H

#ifdef __KERNEL__

#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/lkst_evhandler.h>

extern struct class lkst_class; /* kobject class */

/* new type evhandler interface (it suits for linux 2.6)*/
struct lkst_eh_device {
	/* the name of evhandler */
	char name[LKST_EVHANDLER_NAME_LEN];
	/* the ID of evhandler */
	int id;
	/* event-handler */
	lkst_evhandler_func evhandler;
	lkst_evhandler_ctrl_func evhandler_ctrl;
	/* for evhandler control interface */
	ssize_t (*ctrl_store)(struct lkst_eh_device *, const char *, size_t);
	ssize_t (*ctrl_show)(struct lkst_eh_device *, char *);
	
	struct class_device class_dev;
};

#define to_lkst_eh_device(cd) container_of(cd, struct lkst_eh_device, class_dev);

#define LKST_EH_DEV(name) lkst_eh_dev_##name

#define LKST_EH_DEV_DEF(_name, _evh, _evh_ctl, _cstore, _cshow) \
	struct lkst_eh_device LKST_EH_DEV(_name) = { \
		.name = #_name, \
		.id = LKST_EVHANDLER_ID_VOID, \
		.evhandler = _evh, \
		.evhandler_ctrl = _evh_ctl, \
		.ctrl_store = _cstore, \
		.ctrl_show = _cshow, \
		.class_dev.class = &lkst_class, \
		.class_dev.dev = NULL };

#define LKST_EH_DEV_INIT(ehdev, _name, _evh, _evh_ctl, _cstore, _cshow) \
	({	strncpy(ehdev.name, #_name, LKST_EVHANDLER_NAME_LEN-1 );\
		ehdev.id = LKST_EVHANDLER_ID_VOID; \
		ehdev.evhandler = _evh; \
		ehdev.evhandler_ctrl = _evh_ctl; \
		ehdev.ctrl_store = _cstore; \
		ehdev.ctrl_show = _cshow; \
		ehdev.class_dev.class = &lkst_class; \
		ehdev.class_dev.dev = NULL; \
	})

/* allocate from slab -- may sleep */
extern struct lkst_eh_device * lkst_eh_device_alloc(void);
/* free to slab */
extern void lkst_eh_device_free(struct lkst_eh_device *);

/* register & unregister (including lkst_evhandler_register/unregister)*/
extern int lkst_eh_device_register(struct lkst_eh_device *);
extern void lkst_eh_device_unregister(struct lkst_eh_device *);

#endif /*__KERNEL__*/

#endif
