/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2003-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/
#ifndef _LINUX_LKST_HOOK_H
#define _LINUX_LKST_HOOK_H

#ifdef __KERNEL__
# include <linux/config.h>
# include <linux/types.h>
#endif /*__KERNEL__*/

/* event types enumlation */
#include <linux/lkst_events.h> /* typedef enum{} lkst_event_type;*/

#include <asm/lkst.h>

/* arguments & macros */
typedef u_int64_t lkst_arg_t;

#define LKST_ARG64(val)	((lkst_arg_t)(val))

#define LKST_ARG32(high,low)						\
	((lkst_arg_t) (((lkst_arg_t) (u_int32_t) high << 32) |	\
		           (u_int32_t) low))

#ifdef CONFIG_DEBUG_LKST_DONOT_EXPAND_ARGBITS
#define LKST_ARG(val)	(LKST_ARG32(0,val)) /* do not expand to 64bits */
#else
#define LKST_ARG(val)	(LKST_ARG64(val)) /* expand to 64 bits; 
					   more compatible for IA64. 
					   But larger overhead*/
#endif

#define LKST_ARG32_H(arg) ((u_int32_t)((lkst_arg_t)arg >> 32))
#define LKST_ARG32_L(arg) ((u_int32_t)((lkst_arg_t)arg & \
					       (lkst_arg_t)0x00000000ffffffff))

/* etype flag */
#define LKST_EF_MASK   0xffff
#define LKST_EF_MASKABLE 0x01
#define LKST_EF_MULTIPLE 0x02

/* flag group */
#define LKST_EFG_NORMAL (LKST_EF_MASKABLE)
#define LKST_EFG_INLINE (LKST_EF_MASKABLE | LKST_EF_MULTIPLE)
#define LKST_EFG_UNDEF  (LKST_EF_MULTIPLE)
#define LKST_EFG_MODULE  (LKST_EFG_INLINE)


#ifdef __KERNEL__

# if (defined(CONFIG_DEBUG_KERNEL_STATE_TRACE) ||  \
      defined(CONFIG_DEBUG_KERNEL_STATE_TRACE_MODULE)) && \
      ! defined(__DISABLE_LKST_HOOK__)

# include <linux/hook.h>

/* HOOK & event type Information */
struct lkst_etype_info {
	int etype;
	int flags;
	struct hook * hook;
	struct hook_rec hook_rec;
#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
	char * mnemonic;
	char * name;
	char * args[4];
#endif
};

#define LKST_ETYPE_INFO(_mnemonic) lkst_einfo_##_mnemonic

#ifdef CONFIG_DEBUG_LKST_ETYPES_PROCFS
#define LKST_ETYPE_INFO_DEF(_etype, _eflags, _mnemonic, _name, _args_1, _args_2, _args_3, _args_4, _hook) \
	struct lkst_etype_info LKST_ETYPE_INFO(_mnemonic) = { \
		.etype = _etype, \
		.flags = _eflags, \
		.hook = _hook, \
		.hook_rec = {0}, \
		.mnemonic = #_mnemonic, \
		.name = _name, \
		.args = {_args_1, _args_2, _args_3, _args_4}, \
		};
#else
#define LKST_ETYPE_INFO_DEF(_etype, _eflags, _mnemonic, _name, _args_1, _args_2, _args_3, _args_4, _hook) \
	struct lkst_etype_info LKST_ETYPE_INFO(_mnemonic) = { \
		.etype = _etype, \
		.flags = _eflags, \
		.hook = _hook, \
		.hook_rec = {0}, \
		};
#endif

extern int lkst_hook_etype_register(struct lkst_etype_info *info);
extern int lkst_hook_etype_unregister(struct lkst_etype_info *info);

/* Declaration of hook header. compatible for Kernelhooks. */
#define DECLARE_GENERIC_EXCLUSIVE_HOOK(name) _DECLARE_GENERIC_EXCLUSIVE_HOOK(name, HOOK_SYM(name))
#define _DECLARE_GENERIC_EXCLUSIVE_HOOK(name, hk) \
struct hook name = { \
	.hook_addr 	= NULL, 			  \
	.exit_list	= LIST_HEAD_INIT(name.exit_list), \
	.hook_flags	= HOOK_EXCLUSIVE,		  \
}; \
EXPORT_SYMBOL(name);

#define GENERIC_EXCLUSIVE_HOOK(h, args...) { \
	extern struct hook h; \
	HOOK_DISP_EXCLUSIVE(h, DISPATCH_NORMAL , ##args); \
}

#define LKST_ET_SYM(mnemonic) LKST_ETYPE_##mnemonic
#define LKST_HK_SYM(h) h##_HEADER

/* HOOK Interface */
#define LKST_HOOK(etype, arg1, arg2, arg3, arg4) \
	EXCLUSIVE_HOOK(LKST_HK_SYM(etype), etype, arg1, arg2, arg3, arg4)

#define LKST_HOOK_INLINE(etype, arg1, arg2, arg3, arg4) \
	GENERIC_EXCLUSIVE_HOOK(LKST_HK_SYM(etype), etype, arg1, arg2, arg3, arg4)

/* Declaration of hook header */
#define LKST_HOOK_DECLARE(etype) _LKST_HOOK_DECLARE(LKST_HK_SYM(etype))
#define _LKST_HOOK_DECLARE(ehook) \
	DECLARE_EXCLUSIVE_HOOK(ehook);

#define LKST_HOOK_DECLARE_INLINE(etype) \
	_LKST_HOOK_DECLARE_INLINE(LKST_HK_SYM(etype))
#define _LKST_HOOK_DECLARE_INLINE(ehook) \
	DECLARE_GENERIC_EXCLUSIVE_HOOK(ehook);

#define LKST_ETYPE_DEF_MODULE(id, type, mnemonic, name, arg1, arg2, arg3, arg4) \
	_LKST_ETYPE_DEF_MODULE(id, type, LKST_ET_SYM(mnemonic), mnemonic, name, \
			      arg1, arg2, arg3, arg4)
#define _LKST_ETYPE_DEF_MODULE(id, type, etype, mnemonic, name, arg1, arg2, arg3, arg4) \
	__LKST_ETYPE_DEF_MODULE_##type(id, etype, mnemonic, name, \
				       arg1,arg2,arg3,arg4, \
				       LKST_HK_SYM(etype))

#define __LKST_ETYPE_DEF_MODULE_NORMAL(id, etype, mnemonic, name, arg1,arg2,arg3,arg4,hk) \
enum { etype = id }; \
LKST_HOOK_DECLARE(etype); \
LKST_ETYPE_INFO_DEF(etype, LKST_EFG_NORMAL, mnemonic, name, arg1, arg2, arg3, arg4, &hk);

#define __LKST_ETYPE_DEF_MODULE_INLINE(id, etype, mnemonic, name, arg1,arg2,arg3,arg4,hk) \
enum { etype = id }; \
LKST_HOOK_DECLARE_INLINE(etype); \
LKST_ETYPE_INFO_DEF(etype, LKST_EFG_INLINE, mnemonic, name, arg1, arg2, arg3, arg4, &hk);

#define __LKST_ETYPE_DEF_MODULE_MODULE(id, etype, mnemonic, name, arg1,arg2,arg3,arg4,hk) \
	__LKST_ETYPE_DEF_MODULE_INLINE(id, etype, mnemonic, name, arg1,arg2,arg3,arg4,hk)

#define __LKST_ETYPE_DEF_MODULE_UNDEF(id, etype, mnemonic, name, arg1,arg2,arg3,arg4,hk) \
enum { etype = id }; \
LKST_ETYPE_INFO_DEF(etype, LKST_EFG_UNDEF, mnemonic, name, arg1, arg2, arg3, arg4, NULL);


#else  /* !CONFIG_DEBUG_KERNEL_STATE_TRACE && !CONFIG_DEBUG_KERNEL_STATE_TRACE_MODULE */
/* dummy macros when LKST is not compiled in */
#define LKST_HOOK(etype, arg1, arg2, arg3, arg4)
#define LKST_HOOK_INLINE(etype, arg1, arg2, arg3, arg4)
#define LKST_HOOK_DECLARE(etype)
#define LKST_HOOK_DECLARE_INLINE(etype)
#endif /* CONFIG_DEBUG_KERNEL_STATE_TRACE || CONFIG_DEBUG_KERNEL_STATE_TRACE_MODULE */

#endif /*__KERNEL__*/

#endif
