/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by T.Nakamura <tetsu@crl.hitachi.co.jp>
 * Updated by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _LINUX_LKST_PRIVATE_H
#define _LINUX_LKST_PRIVATE_H

#include <linux/lkst.h>
#include <linux/lkst_maskset.h>
#include <linux/lkst_buffer.h>
#include <linux/lkst_evhandler.h>
#include <asm/lkst_private.h>
#include <asm/page.h>

#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/kernel_stat.h>
#include <linux/string.h>
#include <linux/wait.h>
#include <linux/cache.h>

extern void * lkst_encounter_a_BUG(const char *file, int line);
#define LKST_ENCOUNTER_A_BUG lkst_encounter_a_BUG(__FILE__,__LINE__);

#endif /*__KERNEL__*/

#define LKST_EVHANDLER_BUFFER_TRIAL_MAX		10
#define LKST_EVHANDLER_MASKSET_TRIAL_MAX	10

#define LKST_DEV_MAJOR	231	/* LKST device major number */  /* TODO */
#define LKST_DEV_MINOR	0	/* LKST device minor number */
#define LKST_DEV_NAME	"lkst"	/* LKST device name */

/* LKST buffer version */
#define LKST_BUF_VER	0x000102	/* major(1B) miner(1B) sub(1B) */

#define LKST_MASKSET_TABLE_BYTE	\
	(sizeof(lkst_evhandler_id) * (LKST_ETYPE_MAX + 1))

typedef lkst_evhandler_id * lkst_maskset;

struct lkst_maskset_table {
	char name[LKST_MASKSET_NAME_LEN];
	lkst_maskset * masksetp;
};

struct lkst_event_record {
	int32_t    log_recid;
	int32_t    log_event_type;
	lkst_tsc_t log_time;
	pid_t      log_pid;
  	int32_t    processor;
	lkst_arg_t log_arg1;
	lkst_arg_t log_arg2;
	lkst_arg_t log_arg3;
	lkst_arg_t log_arg4;
	lkst_arg_t _reserved1;
};

struct lkst_evhandler_entry {
	char name[LKST_EVHANDLER_NAME_LEN];
	lkst_evhandler_func evhandler;
	lkst_evhandler_ctrl_func evhandler_ctrl;
};

#define LKST_EVENT_RECORD_BYTE (sizeof(struct lkst_event_record)) /*MUST be 64 bytes*/
#define LKST_LOG_RECORD_BYTE (sizeof(struct lkst_log_record))

#define LKST_BASEID_MASK	0x00ffffff
#define LKST_BUFFER_RECID(bufferid, baseid) \
	(((baseid) & LKST_BASEID_MASK) | ((bufferid) << 24))

/*Only header -- use macros when access to body*/
/* write position is calcurated from recid and initial recid */
struct lkst_event_buffer {
	char eyecatcher[4];			/* EyeCatcher */
	int  buf_ver;				/* LKST buffer version */
	int  cmos_time_flag; 			/* cmos time flag */
	struct timeval xtime;			/* xtime */
	int  cpu;				/* owner processor */
	int  size;				/* buffer size (count). including margin area.*/
	int  baseid;				/* base recid */
	int  init_baseid;			/* initial base recid */
	int  write_offset;			/* write position offset */
	lkst_buffer_id id, next;		/* buffer ID of own/next buffer */
	lkst_buffer_id rq_next, rq_prev;	/* buffer ID of read queue */
};

#define LKST_BUFFER_HEADER_COUNT \
	((int)( (sizeof(struct lkst_event_buffer)-1) \
		   / LKST_EVENT_RECORD_BYTE) + 1)

#define LKST_BUFFER_HEADER_BYTE \
	(LKST_BUFFER_HEADER_COUNT*LKST_EVENT_RECORD_BYTE)

#define LKST_BUFFER_PAGE_BYTE  (PAGE_SIZE)	/* arch dep. */
#define LKST_BUFFER_PAGE_MASK  (~(LKST_BUFFER_PAGE_BYTE - 1))
#define LKST_BUFFER_PAGE_COUNT (LKST_BUFFER_PAGE_BYTE/LKST_EVENT_RECORD_BYTE)

#define LKST_BUFFER_REAL_SIZE(buffer) ((buffer->size*LKST_EVENT_RECORD_BYTE)\
					       + LKST_BUFFER_HEADER_BYTE)

/* Rooms for overrun in buffer wrapping around.	
	 It assumes the worst case that all the CPUs 
	 interrupt in detection of buffer wrapping,
	 and exception occurs on each CPU.		*/
#define LKST_BUFFER_MARGIN_BYTE	  (LKST_BUFFER_PAGE_BYTE-LKST_BUFFER_HEADER_BYTE)
#define LKST_BUFFER_MARGIN_COUNT  (LKST_BUFFER_MARGIN_BYTE/LKST_EVENT_RECORD_BYTE)

#define LKST_BUFFER_INIT_POS	((int)LKST_BUFFER_MARGIN_COUNT)

/* static buffer size -- must be power of 2 */
#if defined(CONFIG_DEBUG_LKST_SBUF_TINY)
# define CONFIG_DEBUG_LKST_SBUF_SIZE 32768
#elif defined(CONFIG_DEBUG_LKST_SBUF_LARGE)
# define CONFIG_DEBUG_LKST_SBUF_SIZE 524288
#else /*defined(CONFIG_DEBUG_LKST_SBUF_MEDIUM)*/
# define CONFIG_DEBUG_LKST_SBUF_SIZE 131072
#endif

#define LKST_STATIC_BUFFER_BYTE		(CONFIG_DEBUG_LKST_SBUF_SIZE)	/* Configurable */
#define LKST_STATIC_BUFFER_COUNT	(LKST_STATIC_BUFFER_BYTE/LKST_EVENT_RECORD_BYTE)
#define LKST_STATIC_BUFFER_MASK		(LKST_STATIC_BUFFER_COUNT - 1)

/* temporary buffer that used as cache to read from user space.*/
#define LKST_TEMP_BUFFER_COUNT		128 /* 128 entries*/
#define LKST_TEMP_BUFFER_BYTE		(LKST_TEMP_BUFFER_COUNT*LKST_LOG_RECORD_BYTE)

#define LKST_OWNER_PID_VOID (PID_MAX_LIMIT + 1)

#if defined(__KERNEL__) || defined( CMD_LKSTDUMP)

#include <asm/atomic.h>

struct lkst_static_buffer {
	struct lkst_event_record body[LKST_STATIC_BUFFER_COUNT]; /* buffer body */
	atomic_t recid;			/* write_pos = recid % LKST_STATIC_BUFFER_COUNT */
} ____cacheline_aligned;

struct lkst_buffer_owner {
	pid_t pid;			/* owner pid */
	lkst_buffer_id read_buf;	/* current read buffer */
	char _pad[3];
	int read_pos;			/* current read position (count)*/
 	int read_baseid;		/* baseid of the event record in read_pos */
	int ready;			/* ready to read flag */
	wait_queue_head_t wait;		/* buffer wait queue */
};

struct lkst_header_percpu {
	int cpu;			/* bound cpu (for lcrash)*/
	lkst_buffer_id write_buf;	/* IDs of current buffers. */
	lkst_buffer_id rq_tail;		/* Tail of read queue(in) */
	lkst_buffer_id rq_head;		/* Head of read queue(out) */
	char _pad;
	int reading;			/* per processor reading flag*/
	struct lkst_buffer_owner owner;
	char eyecatcher[4];		/* EyeCatcher */
	struct {
		struct lkst_event_buffer * ptr;
	} buffer_table[LKST_BUFFER_TBL_MAX];
} ____cacheline_aligned;

//TODO : header size is changed by the architecture.
struct lkst_header {
	char eyecatcher[4];	/* EyeCatcher */
	lkst_maskset_id current_maskset_id;	/* Current maskset id */
	lkst_maskset_id maskset_id_saved;	/* Saved maskset id */
	char _pad[2];
	
	/* Current Maskset Pointer */
	lkst_maskset * current_masksetp;

	/* Maskset Table */
	struct lkst_maskset_table maskset_table[LKST_MASKSET_TBL_MAX];

	/* Event Handler Table */
	struct lkst_evhandler_entry evhandler_table[LKST_EVHANDLER_TBL_MAX];

	/* LKST available memory size */
	unsigned long available_memory_size;
	unsigned long used_memory_size;

	int endian_big;			/* byte order, 0 if little endian */
	lkst_cpu_freq_t	cpu_freq;	/* cpu clockspeed in kHz */
	lkst_tsc_t tsc;			/* machine cycle */
	struct timeval xtime;		/* xtime */
	unsigned int cmos_time_flag; 	/* cmos time flag */
	int buf_ver;			/* LKST buffer version */
	char arch[LKST_ARCH_NAME_LEN];	/* Architecture name */

	struct lkst_static_buffer static_buffer;
};


/* macros to access to the body of buffers */
				/* top address of buffer body */
#define LKST_BUFFER_BODY(buffer) \
	((struct lkst_event_record *)((char *)(buffer)+LKST_BUFFER_HEADER_BYTE))
				/* entry address indexed in buffer body */
#define LKST_BUFFER_BODY_INDEX(buffer,index) (LKST_BUFFER_BODY(buffer)[(index)])

				/* Offset of buffer body from buffer head */
#define LKST_BUFFER_BODY_OFFSET() LKST_BUFFER_HEADER_BYTE

				/* size of the buffer whole */
#define LKST_BUFFER_AREA_SIZE(n)					\
	(LKST_BUFFER_BODY_OFFSET() + (n))

#define LKST_STATIC_BUFFER_BODY() \
	((struct lkst_event_record *)lkst.static_buffer.body)
#define LKST_STATIC_BUFFER_BODY_INDEX(index) (LKST_STATIC_BUFFER_BODY()[(index)])


#endif /*defined(__KERNEL__) || defined( CMD_LKSTDUMP)*/

#ifdef __KERNEL__

extern struct lkst_header lkst;

/* Implemented in lkst_kernel.c */
extern lkst_cpu_freq_t lkst_get_cpu_khz(void);
extern struct lkst_header_percpu lkst_cpu[LKST_CPU_MAX];

				/* header structure of current processor */
#define lkst_current lkst_cpu[smp_processor_id()]

				/* buffer access macros */
#define LKST_BUFFER_LOAD(id, pt) \
	(pt = lkst_current.buffer_table[(id)].ptr)

#define LKST_CURRENT_BUFFER_LOAD(pt) \
	((lkst_current.write_buf != LKST_BUFFER_ID_VOID ) && \
	 LKST_BUFFER_LOAD(lkst_current.write_buf, (pt)) )

#define LKST_BUFFER_STORE(id, pt) \
	(lkst_current.buffer_table[(id)].ptr = (pt))

#define LKST_BUFFER_CHECK_NULL(id) \
	( lkst_current.buffer_table[(id)].ptr == NULL )

/*
 * event handler register (GPL only!)
 */

extern int lkst_evhandler_register(lkst_evhandler_id id, 
				   const char * name,
				   lkst_evhandler_func evhandler,
				   lkst_evhandler_ctrl_func evhandler_ctrl);
extern int lkst_evhandler_unregister(lkst_evhandler_id id);
extern int lkst_evhandler_get_id(const char * name);

/*
 * event handler primitive functions
 */

extern struct lkst_event_buffer * \
lkst_evhandlerprim_select_buffer_id(lkst_buffer_id id);

extern struct lkst_event_buffer * \
lkst_evhandlerprim_select_buffer(void);

extern void lkst_evhandlerprim_owner_wakeup(void);

extern void lkst_evhandlerprim_entry_log(int event_type,
					 lkst_arg_t arg1, lkst_arg_t arg2,
					 lkst_arg_t arg3, lkst_arg_t arg4);
#define lkst_evhandlerprim_entry_log_l2(event_type, arg1, arg2, arg3, arg4) \
	lkst_evhandlerprim_entry_log(event_type,                       \
				     arg1, arg2, arg3, arg4)

extern void lkst_evhandlerprim_entry_log_static(int event_type,
						lkst_arg_t arg1, lkst_arg_t arg2,
						lkst_arg_t arg3, lkst_arg_t arg4);

#define lkst_evhandlerprim_buffer_jump(next) lkst_buffer_jump(next)
#define lkst_evhandlerprim_buffer_shift() lkst_buffer_shift()

/* System default event handlers */

extern void lkst_evhandler_default(void * phookrec, int event_type,
				   lkst_arg_t arg1, lkst_arg_t arg2,
				   lkst_arg_t arg3, lkst_arg_t arg4);
extern void lkst_evhandler_log_static(void * phookrec, int event_type,
				      lkst_arg_t arg1, lkst_arg_t arg2,
				      lkst_arg_t arg3, lkst_arg_t arg4);
extern int lkst_evhandler_default_ctrl(void * buf, size_t buf_size);
extern void lkst_evhandler_buffer_shift(void * phookrec, int event_type,
					lkst_arg_t arg1, lkst_arg_t arg2,
					lkst_arg_t arg3, lkst_arg_t arg4);
extern void lkst_evhandler_buffer_shift_wakeup(void * phookrec, int event_type,
					       lkst_arg_t arg1, lkst_arg_t arg2,
					       lkst_arg_t arg3, lkst_arg_t arg4);

/*
 * This migrates current process to specified CPU
 * and save its cpus_allowed flags.
 * This force to migrate current process to specified cpu.
 */
static __inline__ int lkst_cpu_migrate_save(unsigned long *allowed, int cpu)
{
#ifdef CONFIG_SMP 
	unsigned long new_mask;
	*allowed = current->cpus_allowed;
	new_mask = (( 1UL << cpu ) & cpu_online_map);
	if ( new_mask != 0UL ) {
		set_cpus_allowed(current, new_mask);
		return 0;
	} else {
		return -EINVAL;
	}
	return 0;
#else /*!CONFIG_SMP*/
	return (cpu==0)?0:-EINVAL;
#endif /*CONFIG_SMP*/
}

/*
 * This restore current process's cpus_allowed flags. 
 */
static __inline__ void lkst_cpu_migrate_restore(unsigned long allowed)
{
#ifdef CONFIG_SMP 
	set_cpus_allowed(current, allowed);
#endif /*CONFIG_SMP*/
}

/*
 * This check whether the cpu is online.
 */
static __inline__ int lkst_cpu_is_online(int cpu)
{
#ifdef CONFIG_SMP 
	return cpu_online(cpu);
#else /*!CONFIG_SMP*/
	return (cpu==0)?1:0;
#endif /*CONFIG_SMP*/
}

/* implemented in lkst_kernel.c */
extern int lkst_is_text_address(unsigned long addr);

/*
 * function lkst_evhandlerprim_kernel_text_address
 * A primitive part of an Event-handler. This returns whether
 * addr is in kernel (or kernel modules) text address.
 *
 *  Arguments
 *    addr {IN}: address to investigate.
 *
 *  returns
 *    0: addr is out of the kernel code nor kernel module.
 *    non-0: addr is in the kernel code or kernel module.
 *
 */
static __inline__ int \
lkst_evhandlerprim_kernel_text_address(unsigned long addr) {
	return lkst_is_text_address(addr);
}

#endif /* __KERNEL__ */


#define LKST_ETYPE_VERIFY(event_type) 					\
	(!(0xfffff000&(event_type)))

#define LKST_MASKSET_TABLE_LOAD(i, ptr) \
	(ptr = lkst.maskset_table[(i)].masksetp)

#define	LKST_MASKSET_QUERY_EVHANDLER_ID(maskset_ptr, event_type)	\
	(((lkst_evhandler_id *)(maskset_ptr))[(event_type)])
#define	LKST_MASKSET_SET_EVHANDLER_ID(maskset_ptr, event_type, id)	\
	(((lkst_evhandler_id *)(maskset_ptr))[(event_type)] = (id))

#define	LKST_MASKSET_TABLE_ENTRY_SET(table, id, maskname, addr) ({     	\
	((table)[(id)].masksetp = (lkst_maskset *)(addr));			\
	strncpy((table)[(id)].name, (maskname),				\
		(size_t) LKST_MASKSET_NAME_LEN);			\
	(table)[(id)].name[LKST_MASKSET_NAME_LEN-1] = '\0';})

#define	LKST_MASKSET_SET_CURRENT(id) ({					\
	lkst.current_maskset_id = (id);					\
	(lkst.current_masksetp = lkst.maskset_table[(id)].masksetp);})


#endif /* _LINUX_LKST_PRIVATE_H */
