/*
 * kerntypes.c
 *
 * Copyright (C) 2000 Tom Morano (tjm@sgi.com) and
 *                    Matt D. Robinson (yakker@alacritech.com)
 *
 * Dummy module that includes headers for all kernel types of interest. 
 * The kernel type information is used by the lcrash utility when 
 * analyzing system crash dumps or the live system. Using the type 
 * information for the running system, rather than kernel header files,
 * makes for a more flexible and robust analysis tool.
 *
 * This source code is released under version 2 of the GNU GPL.
 */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#include <linux/module.h>
#include <linux/autoconf.h>
#include <linux/mm.h>
#include <linux/config.h>
#include <linux/utsname.h>
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
#include <linux/dump.h>
#endif
#include <linux/pci.h>

/** 
 ** Various type definitions lifted from kernel/sched.c
 **/

#if CONFIG_SHARE_RUNQUEUE
# define MAX_NR_SIBLINGS CONFIG_NR_SIBLINGS
#else
# define MAX_NR_SIBLINGS 1
#endif

#define BITMAP_SIZE ((((MAX_PRIO+1+7)/8)+sizeof(long)-1)/sizeof(long))

typedef struct runqueue runqueue_t;

struct prio_array {
	int nr_active;
	unsigned long bitmap[BITMAP_SIZE];
	struct list_head queue[MAX_PRIO];
};

typedef struct cpu_s {
	task_t *curr, *idle;
	task_t *migration_thread;
	struct list_head migration_queue;
	int active_balance;
	int cpu;
} cpu_t;

struct runqueue {
	spinlock_t lock;
	unsigned long nr_running, nr_switches, expired_timestamp,
			nr_uninterruptible;
	atomic_t nr_iowait;
	struct mm_struct *prev_mm;
	prio_array_t *active, *expired, arrays[2];
	int prev_cpu_load[NR_CPUS];
	int nr_cpus;
	cpu_t cpu[MAX_NR_SIBLINGS];

} ____cacheline_aligned;

void
kerntypes_dummy(void)
{
}
