/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2004
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.SERIZAWA <serizawa@sdl.hitachi.co.jp>
 * Updated by M.Hiramatsu <hiramatu@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/module.h>
#include <linux/lkst_private.h>
#include <linux/vmalloc.h>
#include <linux/highmem.h>
#include <linux/init.h>

#define BUFFER_TBL_INIT {NULL}
#define MASKSET_TBL_INIT {"", NULL}
#define EVHANDLER_TBL_INIT {"", NULL, NULL}

/* LKST availabe memory size -- max vmalloc size */
//#define LKST_AVAILABLE_MEMORY_SIZE	(unsigned long)((unsigned long)VMALLOC_END - (unsigned long)VMALLOC_START - (unsigned long)PAGE_SIZE)
//#define LKST_AVAILABLE_MEMORY_SIZE	(64<<20)

struct lkst_header_percpu lkst_cpu[LKST_CPU_MAX] = {
	[0 ... (LKST_CPU_MAX - 1)] = {
		.cpu=			0,
		.write_buf=		LKST_BUFFER_ID_VOID,
		.rq_tail=		LKST_BUFFER_ID_VOID,
		.rq_head=		LKST_BUFFER_ID_VOID,
		.reading=		0,
		.owner= 		{
			.pid=		LKST_OWNER_PID_VOID,
			.read_buf=	LKST_BUFFER_ID_VOID,
			.read_pos=	0,
			.ready=		0,
			.wait=		{SPIN_LOCK_UNLOCKED, {0}},
		},
		.eyecatcher=		{'L','K','C','P'},
		.buffer_table=		{[0 ... (LKST_BUFFER_TBL_MAX-1)] = {NULL}},
	},
} ;
EXPORT_SYMBOL_GPL(lkst_cpu);

struct lkst_header lkst = {
	.eyecatcher=		{'L','K','S','T'},
	.current_masksetp=	NULL,
  	.current_maskset_id=	0,
 	.maskset_table= 	{[0 ... (LKST_MASKSET_TBL_MAX-1)] = MASKSET_TBL_INIT},
 	.evhandler_table=	{[0 ... (LKST_EVHANDLER_TBL_MAX-1)] = EVHANDLER_TBL_INIT},
 	.maskset_id_saved=	LKST_MASKSET_ID_RDEFAULT,
	.available_memory_size=	LKST_AVAILABLE_MEMORY_MAX,
	.used_memory_size=	0,
	.endian_big=		LKST_BIG_ENDIAN, /*if arch-endian is big, then 1*/
	.cpu_freq=		0,
	.tsc=			0,
	.xtime=			{0 ,0},
	.cmos_time_flag=	1,
	.buf_ver=		LKST_BUF_VER,
	.arch=			{UTS_MACHINE},
	.static_buffer=		{
		.recid=		ATOMIC_INIT(0),
	}
};
EXPORT_SYMBOL_GPL(lkst); // now, you can access directory.

/* Definitions of hook header */
#define LKST_ETYPE_DEF(event_type, hooktype, mnemonic, event_name, arg_1, arg_2, arg_3, arg_4) \
        LKST_HOOK_DECL_##hooktype(LKST_ET_SYM(mnemonic));

#define LKST_HOOK_DECL_NORMAL(etype) LKST_HOOK_DECLARE(etype)
#define LKST_HOOK_DECL_INLINE(etype) LKST_HOOK_DECLARE_INLINE(etype)
#define LKST_HOOK_DECL_MODULE(etype) LKST_HOOK_DECLARE_INLINE(etype)
#define LKST_HOOK_DECL_UNDEF(etype)

#include <linux/lkst_etype.h>

#undef  LKST_ETYPE_DEF

/*
 * function lkst_time_set
 * This function sets time stamp of LKST header.
 * 
 *  Arguments
 *    from_settimeofday : flag shows to be called from do_sys_settimeofday().
 *  returns
 *    NONE
 * 
 */
void lkst_time_set(int from_settimeofday)
{
	unsigned long flags;
	local_irq_save(flags);
	if (from_settimeofday)
		lkst.cmos_time_flag = 0;
	do_gettimeofday(&lkst.xtime);
	lkst.tsc = lkst_evhandlerprim_mc();
	local_irq_restore(flags);
}
EXPORT_SYMBOL(lkst_time_set);


/*
 * This function is called by lkst modules to get
 * variable "cpu_khz" from LKST_GET_CPU_KHZ().
 * In some arch, LKST will not provide this function.
 */
lkst_cpu_freq_t lkst_get_cpu_khz(void)
{
#ifdef	LKST_GET_CPU_KHZ
	return LKST_GET_CPU_KHZ();
#else
	return 1;
#endif /* ifdef	LKST_GET_CPU_KHZ */
}
EXPORT_SYMBOL(lkst_get_cpu_khz);

extern char _stext, _etext;
/*
 * function lkst_is_text_address
 * Check specified address is contained by text area.
 * 
 *  Caution: This function does not verify value of event_entry_p.
 * 
 *  Arguments
 *    addr {IN}:	address.
 *
 *  returns
 *    if addr is not included in text area, this returns 0.
 *    Otherwise, returns non-0.
 * 
 */
/*
int lkst_is_text_address(unsigned long addr)
{
	if (addr >= (unsigned long)_stext &&
	    addr <= (unsigned long)_etext)
		return 1;//in kernel
	return (module_text_address(addr)!=NULL);
}
EXPORT_SYMBOL(lkst_is_text_address);
*/

#ifdef CONFIG_MODULES
/* These symbols are not exported. so we must
   embed this function into kernel itself. */
extern struct module *module_list;
extern struct module kernel_module;
/*
 * function lkst_find_module_by_address
 * Find module in which specified address is contained, and return
 * pointer to the module.
 *
 *  Note: This function is used from stack trace module only.
 *
 *  Arguments
 *    addr {IN}:        address.
 *
 *  returns
 *    address to the module.
 *
 */
static __inline__ struct module * \
lkst_find_module_by_address_inline(unsigned long addr)
{
	struct module *mod;
	
	if (addr >= (unsigned long) &_stext &&
	    addr <= (unsigned long) &_etext)
		return &kernel_module;
	
	// TODO: avoid conflicts without spin_lock(&modlist_lock);
	
	for (mod = module_list;
             mod != &kernel_module &&
		     mod != NULL ; mod = mod->next) {
		if (!(mod->flags&MOD_RUNNING))
			return NULL;
		if (mod->flags&MOD_DELETED)
			return NULL;
		/* mod_bound tests for addr being inside the vmalloc'ed
		 * module area. Of course it'd be better to test only
		 * for the .text subset... */
		if (mod_bound(addr, 0, mod)) {
			return mod;
			break;
		}
	}
	/* address was not found. */
	return NULL;
}
#else /* !CONFIG_MODULES */
static __inline__ struct module * \
lkst_find_module_by_address_inline(unsigned long addr)
{
        if (addr >= (unsigned long) &_stext &&
            addr <= (unsigned long) &_etext)
                return &kernel_module;
        return NULL;
}
#endif /* CONFIG_MODULES */

/*
 * function lkst_find_module_by_address
 * Find module in which specified address is contained, and return
 * pointer to the module.
 *
 *  Caution: This function does not verify value of event_entry_p.
 *
 *  Arguments
 *    addr {IN}:        address.
 *
 *  returns
 *    address to the module.
 *
 */

struct module * lkst_find_module_by_address(unsigned long addr)
{
        return (lkst_find_module_by_address_inline(addr));
}
EXPORT_SYMBOL(lkst_find_module_by_address);


/*
 * Function lkst_maxvm_setup
 * This function gets size of lkst available memory area from kernel parameter.
 */

static int __init lkst_maxvm_setup(char *line)
{
	int size;
	if (get_option(&line, &size)) {
		if ( size < LKST_AVAILABLE_MEMORY_MIN ) {
			size = LKST_AVAILABLE_MEMORY_MIN;
		}

		if ( size > LKST_AVAILABLE_MEMORY_MAX ) {
			size = LKST_AVAILABLE_MEMORY_MAX;
		}

		lkst.available_memory_size = size;
	}
	return 1;
}
 
__setup("lkst_maxvm=", lkst_maxvm_setup);

/* instead of module param. */
#ifdef CONFIG_DEBUG_KERNEL_STATE_TRACE

extern int lkst_init_maskset;

static int __init lkst_startup_maskset(char *line)
{
	int mask;
	if (get_option(&line, &mask)) {
		if (mask >= 0 && mask <=2)
			lkst_init_maskset = (lkst_maskset_id) mask;
	}
	return 1;
}	
 
__setup("lkst_maskset=", lkst_startup_maskset);

#endif /* CONFIG_DEBUG_KERNEL_STATE_TRACE*/

